<img src="https://github.com/mitsuba-renderer/mitsuba2/raw/master/docs/images/logo_plain.png" width="120" height="120" alt="Mitsuba logo">


Radiative Backpropagation - Initial release
===========================================

This is an initial release of the Radiative Backpropagation implementation (August 2020).

> Nimier-David, Merlin, et al. "Radiative backpropagation: an adjoint method for lightning-fast differentiable rendering."
> ACM Transactions on Graphics (TOG) 39.4 (2020): 146-1.


**Disclaimer**: this is unsupported experimental research code. It is based on an older version of Mitsuba 2, which uses OptiX 6.
As of August 2020, we are working on a more polished implementation based on the latest Mitsuba 2 version and using OptiX 7.


Structure
---------

In order to run the provided examples, the scene data should also be downloaded and placed alongside of the code directory as follows:

```
radiative-backprop
├── scenes
├── mitsuba2
│   ├── build
│   ├── CMakeLists.txt
│   └── ...
└── outputs   // Will be created automatically when running optimizations
```


Building
--------

This was tested on an Ubuntu 20.04 machine with Clang 9.0.1, CUDA 11 and OptiX 6.5 with an NVIDIA RTX 2080Ti GPU (11GB of GPU RAM).

OptiX 6.5 should be installed to `/opt/optix`, or the CMake option `MTS_OPTIX_PATH` should be set to the install location.

```
cd mitsuba2
mkdir build && cd build
cmake .. -GNinja -DMTS_USE_EMBREE=ON -DMTS_ENABLE_AUTODIFF=ON
ninja
source ../setpath.sh
```

Running
-------

Optimization configurations are defined in file `mitsuba2/src/optix/tests/config.py`. They refer to a Mitsuba 2 scene file, a reference image, and define several options such as the sample count, etc.

To run one of those optimizations:

```
cd mitsuba2/build
ninja
source ../setpath.sh

python3 ../src/optix/tests/optimize_rb.py <config_name>
```

with `<config_name>` one of the keys defined `config.py`. Scenes used in the paper include:

- `bell`
- `smoke-plume-res-111-white`
- `albedo-opt-null`

Note that some configurations are simply left over from early tests, and as such may not run.
To enable biased variants of radiative backpropagation described in Section 3.8 of the paper, add either or both of the following flags to the script:

```
--use_approx_li_1   # Biased I
--use_interleaving  # Biased II
```

Known issues
------------

**GPU memory usage**: the configuration parameters (e.g. sample per pixel count when rendering reference images) were set on a system with 11GB of GPU RAM. If you experinece crashes related to memory usage, adapt the parameters to reduce the wavefront size so that it fits in memory.


**Multi-GPU systems**: multi-GPU setups are not supported. Please set the following environment variables when running, adapting it with the index of the GPU you would like to use:

```
export CUDA_VISIBLE_DEVICES=1
```

**Occasional deadlock**: we have occasionally encountered a deadlock, which unfortunately we were not able to resolve. It seems to occur (rarely) when calling into OptiX, and is characterized by zero CPU & GPU utilization.
In case you find a run stuck in such a way, the process should be killed and started again.

**CUDA build errors when building Enoki**: build errors such as `error: identifier "__builtin_ia32_sqrtsd_round" is undefined` [have been reported](https://github.com/mitsuba-renderer/mitsuba2/issues/27). In that case, remove the `-march=native` flag in `ext/enoki/CMakeLists.txt`.

**OptiX not found**: make sure that the path to the OptiX library is in the search path (e.g. using `LD_LIBRARY_PATH`).

---

Below are general Mitsuba 2-related instructions:


# Mitsuba Renderer 2

Mitsuba 2 is a research-oriented rendering system written in portable C++17.
It consists of a small set of core libraries and a wide variety of plugins that
implement functionality ranging from materials and light sources to complete
rendering algorithms.

## Compiling

Compiling from scratch requires CMake and a recent version of XCode on Mac,
Visual Studio 2017 on Windows, or GCC/Clang (Clang is preferred) on Linux.

On Linux and MacOS, compiling should be as simple as

```bash
git clone --recursive https://github.com/mitsuba-renderer/mitsuba2
cd mitsuba2

# Build using CMake & GNU Make (legacy)
cmake .
make -j 4

# OR: Build using CMake & Ninja (recommended, must install Ninja first)
cmake -GNinja .
ninja
```

On Windows, open the generated ``mitsuba.sln`` file after running
``cmake .`` and proceed building as usual from within Visual Studio.

## Running Mitsuba

Once Mitsuba is compiled, run the ``setpath.sh/bat`` script to configure
environment variables (``PATH/LD_LIBRARY_PATH/PYTHONPATH``) that are
required to run Mitsuba.

```bash
# On Linux / Mac OS
$ source setpath.sh

# On Windows
C:\...\mitsuba2> setpath
```

## Running the tests

To run the test suite, simply invoke ``pytest``:

```bash
$ pytest
```

The build system also exposes a ``pytest`` target that executes ``setpath`` and
parallelizes the test execution.

```bash
# if using Makefiles
$ make pytest

# if using ninja
$ ninja pytest
```

## Staying up-to-date

Mitsuba organizes its software dependencies in a hierarchy of sub-repositories
using *git submodule*. Unfortunately, as of 2018, pulling from the main
repository won't automatically keep the sub-repositories in sync, which can
lead to various problems. The following command installs a git alias named
``pullall`` that automates these two steps.

```bash
git config --global alias.pullall '!f(){ git pull "$@" && git submodule update --init --recursive; }; f'
```

Afterwards, simply write
```bash
git pullall
```
to stay in sync.

## About

This project was created by [Wenzel Jakob](http://rgl.epfl.ch/people/wjakob)
Significant features and/or improvements to the code were contributed by
Merlin Nimier-David and Sébastien Speierer.
