// -----------------------------------------------------------------------
//! @{ \name Macro for template implementation of endpoint methods
// -----------------------------------------------------------------------

#define MTS_IMPLEMENT_ENDPOINT_SCALAR()                                        \
    using Endpoint::sample_ray;                                                \
    using Endpoint::sample_direction;                                          \
    using Endpoint::sample_position;                                           \
    using Endpoint::sample_wavelengths;                                        \
    using Endpoint::pdf_direction;                                             \
    using Endpoint::pdf_position;                                              \
    using Endpoint::pdf_wavelengths;                                           \
    using Endpoint::eval;                                                      \
                                                                               \
    std::pair<Ray3f, Spectrumf> sample_ray(                                    \
        Float time, const Spectrumf &wavelengths, const Point2f &sample1,      \
        const Point2f &sample2) const override {                               \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRay);                     \
        return sample_ray_impl(time, wavelengths, sample1, sample2, true);     \
    }                                                                          \
    std::pair<DirectionSample3f, Spectrumf> sample_direction(                  \
        const Interaction3f &ref, const Point2f &sample) const override {      \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirection);               \
        return sample_direction_impl(ref, sample, true);                       \
    }                                                                          \
    std::pair<PositionSample3f, Float> sample_position(                        \
        Float time, const Point2f &sample) const override {                    \
        return sample_position_impl(time, sample, true);                       \
    }                                                                          \
    std::pair<Spectrumf, Spectrumf> sample_wavelengths(Float sample)           \
        const override {                                                       \
        return sample_wavelengths_impl(sample, true);                          \
    }                                                                          \
    Float pdf_direction(const Interaction3f &ref, const DirectionSample3f &ds) \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return pdf_direction_impl(ref, ds, true);                              \
    }                                                                          \
    Float pdf_position(const PositionSample3f &ps) const override {            \
        return pdf_position_impl(ps, true);                                    \
    }                                                                          \
    Spectrumf pdf_wavelengths(const Spectrumf &wavelengths) const override {   \
        return pdf_wavelengths_impl(wavelengths, true);                        \
    }                                                                          \
    Spectrumf eval(const SurfaceInteraction3f &si) const override {            \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return eval_impl(si, true);                                            \
    }

#define MTS_IMPLEMENT_ENDPOINT_PACKET()                                        \
    std::pair<Ray3fP, SpectrumfP> sample_ray(                                  \
        FloatP time, const SpectrumfP &wavelengths, const Point2fP &sample1,   \
        const Point2fP &sample2, MaskP active) const override {                \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRayP);                    \
        return sample_ray_impl(time, wavelengths, sample1, sample2, active);       \
    }                                                                          \
    std::pair<DirectionSample3fP, SpectrumfP> sample_direction(                \
        const Interaction3fP &ref, const Point2fP &sample, MaskP active)       \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirectionP);              \
        return sample_direction_impl(ref, sample, active);                     \
    }                                                                          \
    std::pair<PositionSample3fP, FloatP> sample_position(                      \
        FloatP time, const Point2fP &sample, MaskP active) const override {    \
        return sample_position_impl(time, sample, active);                     \
    }                                                                          \
    std::pair<SpectrumfP, SpectrumfP> sample_wavelengths(                      \
        const FloatP &sample, MaskP active) const override {                   \
        return sample_wavelengths_impl(sample, active);                        \
    }                                                                          \
    FloatP pdf_direction(const Interaction3fP &ref,                            \
                         const DirectionSample3fP &ds, MaskP active)           \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluateP);                     \
        return pdf_direction_impl(ref, ds, active);                            \
    }                                                                          \
    FloatP pdf_position(const PositionSample3fP &ps, MaskP active)             \
        const override {                                                       \
        return pdf_position_impl(ps, active);                                  \
    }                                                                          \
    SpectrumfP pdf_wavelengths(const SpectrumfP &wavelengths, MaskP active)    \
        const override {                                                       \
        return pdf_wavelengths_impl(wavelengths, active);                      \
    }                                                                          \
    SpectrumfP eval(const SurfaceInteraction3fP &si, MaskP active)             \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluateP);                     \
        return eval_impl(si, active);                                          \
    }

#if !defined(MTS_ENABLE_AUTODIFF)
#  define MTS_IMPLEMENT_ENDPOINT_AUTODIFF()
#else
#define MTS_IMPLEMENT_ENDPOINT_AUTODIFF()                                      \
    std::pair<Ray3fD, SpectrumfD> sample_ray(                                  \
        FloatD time, const SpectrumfD &wavelengths, const Point2fD &sample1,   \
        const Point2fD &sample2, MaskD active) const override {                \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRay);                     \
        return sample_ray_impl(time, wavelengths, sample1, sample2, active);   \
    }                                                                          \
    std::pair<DirectionSample3fD, SpectrumfD> sample_direction(                \
        const Interaction3fD &ref, const Point2fD &sample, MaskD active)       \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirection);               \
        return sample_direction_impl(ref, sample, active);                     \
    }                                                                          \
    std::pair<PositionSample3fD, FloatD> sample_position(                      \
        FloatD time, const Point2fD &sample, MaskD active) const override {    \
        return sample_position_impl(time, sample, active);                     \
    }                                                                          \
    std::pair<SpectrumfD, SpectrumfD> sample_wavelengths(                      \
        const FloatD &sample, MaskD active) const override {                   \
        return sample_wavelengths_impl(sample, active);                        \
    }                                                                          \
    FloatD pdf_direction(const Interaction3fD &ref,                            \
                         const DirectionSample3fD &ds, MaskD active)           \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return pdf_direction_impl(ref, ds, active);                            \
    }                                                                          \
    FloatD pdf_position(const PositionSample3fD &ps, MaskD active)             \
        const override {                                                       \
        return pdf_position_impl(ps, active);                                  \
    }                                                                          \
    SpectrumfD pdf_wavelengths(const SpectrumfD &wavelengths, MaskD active)    \
        const override {                                                       \
        return pdf_wavelengths_impl(wavelengths, active);                      \
    }                                                                          \
    SpectrumfD eval(const SurfaceInteraction3fD &si, MaskD active)             \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return eval_impl(si, active);                                          \
    }
#endif

#define MTS_IMPLEMENT_ENDPOINT_ALL()                                           \
    MTS_IMPLEMENT_ENDPOINT_SCALAR()                                            \
    MTS_IMPLEMENT_ENDPOINT_PACKET()                                            \
    MTS_IMPLEMENT_ENDPOINT_AUTODIFF()

#define MTS_IMPLEMENT_ENDPOINT_POLARIZED_SCALAR()                              \
    std::pair<Ray3f, MuellerMatrixSf> sample_ray_pol(                          \
        Float time, Float sample1, const Point2f &sample2,                     \
        const Point2f &sample3) const override {                               \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRay);                     \
        return sample_ray_pol_impl(time, sample1, sample2, sample3, true);     \
    }                                                                          \
    std::pair<DirectionSample3f, MuellerMatrixSf> sample_direction_pol(        \
        const Interaction3f &it, const Point2f &sample) const override {       \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirection);               \
        return sample_direction_pol_impl(it, sample, true);                    \
    }                                                                          \
    MuellerMatrixSf eval_pol(const SurfaceInteraction3f &si) const override {  \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return eval_pol_impl(si, true);                                        \
    }

#define MTS_IMPLEMENT_ENDPOINT_POLARIZED_PACKET()                              \
    std::pair<Ray3fP, MuellerMatrixSfP> sample_ray_pol(                        \
        FloatP time, FloatP sample1, const Point2fP &sample2,                  \
        const Point2fP &sample3, MaskP active) const override {                \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRay);                     \
        return sample_ray_pol_impl(time, sample1, sample2, sample3, active);   \
    }                                                                          \
    std::pair<DirectionSample3fP, MuellerMatrixSfP> sample_direction_pol(      \
        const Interaction3fP &it, const Point2fP &sample, MaskP active)        \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirection);               \
        return sample_direction_pol_impl(it, sample, active);                  \
    }                                                                          \
    MuellerMatrixSfP eval_pol(const SurfaceInteraction3fP &si, MaskP active)   \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluateP);                     \
        return eval_pol_impl(si, active);                                      \
    }

#if !defined(MTS_ENABLE_AUTODIFF)
#  define MTS_IMPLEMENT_ENDPOINT_POLARIZED_AUTODIFF()
#else
#  define MTS_IMPLEMENT_ENDPOINT_POLARIZED_AUTODIFF()                          \
    std::pair<Ray3fD, MuellerMatrixSfD> sample_ray_pol(                        \
        FloatP time, FloatP sample1, const Point2fD &sample2,                  \
        const Point2fD &sample3, MaskD active) const override {                \
        ScopedPhase p(EProfilerPhase::EEndpointSampleRay);                     \
        return sample_ray_pol_impl(time, sample1, sample2, sample3, active);   \
    }                                                                          \
    std::pair<DirectionSample3fD, MuellerMatrixSfD> sample_direction_pol(      \
        const Interaction3fD &it, const Point2fD &sample, MaskD active)        \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointSampleDirection);               \
        return sample_direction_pol_impl(it, sample, active);                  \
    }                                                                          \
    MuellerMatrixSfD eval_pol(const SurfaceInteraction3fD &si, MaskD active)   \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EEndpointEvaluate);                      \
        return eval_pol_impl(si, active);                                      \
    }
#endif

#define MTS_IMPLEMENT_ENDPOINT_POLARIZED_ALL()                                 \
    MTS_IMPLEMENT_ENDPOINT_POLARIZED_SCALAR()                                  \
    MTS_IMPLEMENT_ENDPOINT_POLARIZED_PACKET()                                  \
    MTS_IMPLEMENT_ENDPOINT_POLARIZED_AUTODIFF()

//! @}
// -----------------------------------------------------------------------
