// -----------------------------------------------------------------------
//! @{ \name Macro for template implementation of integrator methods
// -----------------------------------------------------------------------

#define MTS_IMPLEMENT_INTEGRATOR_SCALAR()                                      \
    using SamplingIntegrator::eval;                                            \
    Spectrumf eval(const RayDifferential3f &ray, RadianceSample3f &rs)         \
        const override {                                                       \
        return eval_impl(ray, rs, true);                                       \
    }

#define MTS_IMPLEMENT_INTEGRATOR_PACKET()                                      \
    SpectrumfP eval(const RayDifferential3fP &ray, RadianceSample3fP &rs,      \
                    MaskP active) const override {                             \
        return eval_impl(ray, rs, active);                                     \
    }

#if !defined(MTS_ENABLE_AUTODIFF)
#  define MTS_IMPLEMENT_INTEGRATOR_AUTODIFF()
#else
#  define MTS_IMPLEMENT_INTEGRATOR_AUTODIFF()                                  \
    SpectrumfD eval(const RayDifferential3fD &ray, RadianceSample3fD &rs,      \
                    MaskD active) const override {                             \
        return eval_impl(ray, rs, active);                                     \
    }
#endif

#define MTS_IMPLEMENT_INTEGRATOR_ALL()                                         \
    MTS_IMPLEMENT_INTEGRATOR_SCALAR()                                          \
    MTS_IMPLEMENT_INTEGRATOR_PACKET()                                          \
    MTS_IMPLEMENT_INTEGRATOR_AUTODIFF()

// -----------------------------------------------------------------------------

#if !defined(MTS_ENABLE_AUTODIFF)
#  define MTS_IMPLEMENT_LIGHT_TRACER_INTEGRATOR_AUTODIFF()
#else
#define MTS_IMPLEMENT_LIGHT_TRACER_INTEGRATOR_AUTODIFF()                       \
    std::pair<Ray3fD, SpectrumfD> prepare_ray(                                 \
        RadianceSample3fD &rs, const MaskD &active) const override {           \
        return prepare_ray_impl(rs, active);                                   \
    }                                                                          \
    SpectrumfD trace_light_ray(Ray3fD ray, RadianceSample3fD &rs,              \
                               SpectrumfD throughput, ImageBlock *block,       \
                               const MaskD &active) const override {           \
        return trace_light_ray_impl(ray, rs, throughput, block, active);       \
    }                                                                          \
    SpectrumfD connect_sensor(RadianceSample3fD &rs,                           \
                              SurfaceInteraction3fD &si, BSDFPtrD bsdf,        \
                              const SpectrumfD &weight, ImageBlock *block,     \
                              const MaskD &active) const override {            \
        return connect_sensor_impl(rs, si, bsdf, weight, block, active);       \
    }
#endif

// SpectrumfD sample_visible_emitters(RadianceSample3fD &rs, ImageBlock *block,
//                                            MaskD active) const override {
//     return sample_visible_emitters_impl(rs, block, active);
// }
// Spectrumf sample_visible_emitters(RadianceSample3f &rs, ImageBlock *block) \
//     const override {                                                       \
//     return sample_visible_emitters_impl(rs, block, true);                  \
// }                                                                          \
// SpectrumfP sample_visible_emitters(RadianceSample3fP &rs,                  \
//                                    ImageBlock *block, MaskP active = true) \
//     const override {                                                       \
//     return sample_visible_emitters_impl(rs, block, active);                \
// }                                                                          \

/// Instantiates concrete scalar and packet versions of the light integrator API
#define MTS_IMPLEMENT_LIGHT_TRACER_INTEGRATOR()                                \
    using LightTracerIntegrator::sample_visible_emitters;                      \
    using LightTracerIntegrator::prepare_ray;                                  \
    using LightTracerIntegrator::trace_light_ray;                              \
    using LightTracerIntegrator::connect_sensor;                               \
                                                                               \
    std::pair<Ray3f, Spectrumf> prepare_ray(RadianceSample3f &rs)              \
        const override {                                                       \
        return prepare_ray_impl(rs, true);                                     \
    }                                                                          \
    std::pair<Ray3fP, SpectrumfP> prepare_ray(                                 \
        RadianceSample3fP &rs, const MaskP &active) const override {           \
        return prepare_ray_impl(rs, active);                                   \
    }                                                                          \
    Spectrumf trace_light_ray(Ray3f ray, RadianceSample3f &rs,                 \
                              Spectrumf throughput, ImageBlock *block)         \
        const override {                                                       \
        return trace_light_ray_impl(ray, rs, throughput, block, true);         \
    }                                                                          \
    SpectrumfP trace_light_ray(Ray3fP ray, RadianceSample3fP &rs,              \
                               SpectrumfP throughput, ImageBlock *block,       \
                               MaskP active) const override {                  \
        return trace_light_ray_impl(ray, rs, throughput, block, active);       \
    }                                                                          \
    Spectrumf connect_sensor(RadianceSample3f &rs, SurfaceInteraction3f &si,   \
                             const BSDF *bsdf, const Spectrumf &weight,        \
                             ImageBlock *block) const override {               \
        return connect_sensor_impl(rs, si, bsdf, weight, block, true);         \
    }                                                                          \
    SpectrumfP connect_sensor(RadianceSample3fP &rs,                           \
                              SurfaceInteraction3fP &si, BSDFPtrP bsdf,        \
                              const SpectrumfP &weight, ImageBlock *block,     \
                              const MaskP &active) const override {            \
        return connect_sensor_impl(rs, si, bsdf, weight, block, active);       \
    }                                                                          \
    MTS_IMPLEMENT_LIGHT_TRACER_INTEGRATOR_AUTODIFF()

//! @}
// -----------------------------------------------------------------------
