// -----------------------------------------------------------------------
//! @{ \name Enoki support for packets of PhaseFunction pointers
// -----------------------------------------------------------------------

ENOKI_CALL_SUPPORT_BEGIN(mitsuba::PhaseFunction)
    ENOKI_CALL_SUPPORT_METHOD(sample)
    ENOKI_CALL_SUPPORT_METHOD(eval)
ENOKI_CALL_SUPPORT_END(mitsuba::PhaseFunction)

//! @}
// -----------------------------------------------------------------------

/*
 * \brief This macro should be used in the definition of BSDF
 * plugins to instantiate concrete versions of the \c sample,
 * \c eval and \c pdf functions.
 */
#define MTS_IMPLEMENT_PHASE_FUNCTION_SCALAR()                                  \
    using PhaseFunction::sample;                                               \
    using PhaseFunction::eval;                                                 \
    Vector3f sample(const MediumInteraction3f &mi, const Point2f &sample)      \
        const override {                                                       \
        ScopedPhase p(EProfilerPhase::EPhaseFunctionSample);                   \
        return sample_impl(mi, sample, true);                                  \
    }                                                                          \
    Float eval(const Vector3f &wo) const override {                            \
        ScopedPhase p(EProfilerPhase::EBSDFEvaluate);                          \
        return eval_impl(wo, true);                                            \
    }

#define MTS_IMPLEMENT_PHASE_FUNCTION_PACKET()                                  \
    Vector3fP sample(const MediumInteraction3fP &mi, const Point2fP &sample,   \
                     MaskP active = true) const override {                     \
        ScopedPhase p(EProfilerPhase::EPhaseFunctionSampleP);                  \
        return sample_impl(mi, sample, active);                                \
    }                                                                          \
    FloatP eval(const Vector3fP &wo, MaskP active) const override {            \
        ScopedPhase p(EProfilerPhase::EBSDFEvaluateP);                         \
        return eval_impl(wo, active);                                          \
    }

#if !defined(MTS_ENABLE_AUTODIFF)
#define MTS_IMPLEMENT_PHASE_FUNCTION_AUTODIFF()
#else
#define MTS_IMPLEMENT_PHASE_FUNCTION_AUTODIFF()                                \
                                                                               \
    Vector3fD sample(const MediumInteraction3fD &mi, const Point2fD &sample,   \
                     MaskD active = true) const override {                     \
        ScopedPhase p(EProfilerPhase::EPhaseFunctionSampleP);                  \
        return sample_impl(mi, sample, active);                                \
    }                                                                          \
    FloatD eval(const Vector3fD &wo, MaskD active) const override {            \
        ScopedPhase p(EProfilerPhase::EBSDFEvaluateP);                         \
        return eval_impl(wo, active);                                          \
    }
#endif

#define MTS_IMPLEMENT_PHASE_FUNCTION_ALL()                                     \
    MTS_IMPLEMENT_PHASE_FUNCTION_SCALAR()                                      \
    MTS_IMPLEMENT_PHASE_FUNCTION_PACKET()                                      \
    MTS_IMPLEMENT_PHASE_FUNCTION_AUTODIFF()
