#!/bin/bash
#
# Author: Khaja Minhajuddin
# Adapted by: Merlin Nimier-David
#
# File name: git-archive-all
# Usage: cd mitsuba2; ./resources/git-archive-all

set -e
set -C # noclobber

echo "> creating root archive"
export ROOT_ARCHIVE_DIR="$(pwd)"
# create root archive
git archive --verbose --prefix "mitsuba2/" --format "tar" --output "$ROOT_ARCHIVE_DIR/mitsuba2.tar" HEAD

echo "> appending submodule archives"
# For each git submodule (recursively), creates a .tar
# archive with the contents of that submodule, prefixed by
# the right directory structure. Name is chosen to avoid
# conflicts even when the same submodule appears several
# times in the tree.
git submodule foreach --recursive 'safename=$(echo $displaypath | sed -e '"'s/[^A-Za-z0-9._-]/_/g'"');git archive --verbose --prefix="mitsuba2/$displaypath/" --format tar HEAD --output $ROOT_ARCHIVE_DIR/repo-output-sub-$sha1-$safename.tar'

if [[ $(ls repo-output-sub*.tar | wc -l) != 0  ]]; then
  # Combine all archives into one tar
  # Note: can't just use glob because of a bug in tar --concatenate
  # https://superuser.com/a/1098273/439433
  echo "> combining all tars"
  for f in $(ls repo-output-sub*.tar); do
    tar --concatenate --file mitsuba2.tar $f;
  done

  # remove sub tars
  echo "> removing all sub tars"
  rm -rf repo-output-sub*.tar
fi

# gzip the tar
echo "> gzipping final tar"
gzip --force --verbose mitsuba2.tar

# echo "> moving output file to $OUTPUT_FILE"
# mv mitsuba2.tar.gz $OUTPUT_FILE

echo "> git-archive-all done"
