#include <mitsuba/core/properties.h>
#include <mitsuba/core/string.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/render/spectrum.h>

NAMESPACE_BEGIN(mitsuba)

class Null final : public BSDF {
public:
    Null(const Properties &props) : BSDF(props) {
        m_components.push_back(ENull | EFrontSide | EBackSide);
        m_flags = m_components.back();
    }

    template <
        typename SurfaceInteraction, typename Value, typename Point2,
        typename BSDFSample = BSDFSample<typename SurfaceInteraction::Point3>,
        typename Spectrum   = Spectrum<Value>>
    MTS_INLINE std::pair<BSDFSample, Spectrum>
    sample_impl(const BSDFContext &ctx, const SurfaceInteraction &si,
                Value /* unused */, const Point2 & /* unused */,
                mask_t<Value> /* unused */) const {
        bool sample_transmission = ctx.is_enabled(ENull, 0);
        BSDFSample bs;
        Spectrum result(0.f);
        if (sample_transmission) {
            bs.wo                = -si.wi;
            bs.sampled_component = 0;
            bs.sampled_type      = ENull;
            bs.eta               = 1.f;
            bs.pdf               = 1.f;
            result               = 1.f;
        }
        return { bs, result };
    }

    template <typename SurfaceInteraction, typename Vector3,
              typename Value    = typename SurfaceInteraction::Value,
              typename Spectrum = Spectrum<Value>>
    MTS_INLINE Spectrum eval_impl(const BSDFContext & /* unused */,
                                  const SurfaceInteraction & /* unused */,
                                  const Vector3 & /* unused */,
                                  mask_t<Value> /* unused */) const {
        return 0.0f;
    }

    template <typename SurfaceInteraction, typename Vector3,
              typename Value = value_t<Vector3>>
    MTS_INLINE Value pdf_impl(const BSDFContext & /* unused */,
                              const SurfaceInteraction & /* unused */,
                              const Vector3 & /* unused */,
                              mask_t<Value> /* unused */) const {
        return 0.0f;
    }

    template <typename SurfaceInteraction, typename Vector3,
              typename Value    = typename SurfaceInteraction::Value,
              typename Spectrum = Spectrum<Value>>
    MTS_INLINE Spectrum eval_transmission_impl(const SurfaceInteraction & /* unused */,
                                  const Vector3 & /* unused */,
                                  mask_t<Value> /* unused */) const {
        return 1.0f;
    }

    std::string to_string() const override { return "Null[]"; }

    MTS_IMPLEMENT_BSDF_ALL()
    MTS_IMPLEMENT_BSDF_EVAL_TRANSMISSION_ALL()
    MTS_DECLARE_CLASS()
protected:
};

MTS_IMPLEMENT_CLASS(Null, BSDF)
MTS_EXPORT_PLUGIN(Null, "Null material")

NAMESPACE_END(mitsuba)
