#include <mitsuba/core/plugin.h>
#include <mitsuba/core/properties.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/render/emitter.h>
#include <mitsuba/render/endpoint.h>
#include <mitsuba/render/kdtree.h>
#include <mitsuba/render/scene.h>

NAMESPACE_BEGIN(mitsuba)

Emitter::Emitter(const Properties &props) : Endpoint(props) {}

Emitter::~Emitter() {}

bool Emitter::is_environment() const { return false; }

// -----------------------------------------------------------------------------

InfiniteEmitter::InfiniteEmitter(const Properties &props)
    : Emitter(props)
#if defined(MTS_ENABLE_AUTODIFF)
      , m_this_pointer(this_pointer<FloatD>())
#endif
      {
    /* Until `create_shape` is called, we have no information
       about the scene and default to the unit bounding sphere. */
    m_bsphere = BoundingSphere3f(Point3f(0.f), 1.f);
    m_inv_surface_area = rcp(4.f * math::Pi * m_bsphere.radius * m_bsphere.radius);
}

bool InfiniteEmitter::is_environment() const { return true; }

ref<Shape> InfiniteEmitter::create_shape(const Scene *scene) {
    // Create a bounding sphere that surrounds the scene
    auto bbox = scene->bbox();
    m_bsphere          = bbox.bounding_sphere();
    m_bsphere.radius   = max(math::Epsilon, m_bsphere.radius * 1.5f);
    m_inv_surface_area = rcp(4.f * math::Pi * m_bsphere.radius * m_bsphere.radius);

    Properties props("sphere");
    props.set_point3f("center", m_bsphere.center);
    props.set_float("radius", m_bsphere.radius);
    props.set_int("color_mode", m_color_mode);
    props.mark_queried("color_mode");

    // Sphere is "looking in" towards the scene
    props.set_bool("flip_normals", true);

    return PluginManager::instance()->create_object<Shape>(props);
}

MTS_IMPLEMENT_CLASS(Emitter, Endpoint)
MTS_IMPLEMENT_CLASS(InfiniteEmitter, Emitter)
NAMESPACE_END(mitsuba)
