#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/spectrum.h>
#include <mitsuba/core/transform.h>
#include <mitsuba/python/python.h>
#include <mitsuba/render/endpoint.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/records.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/shape.h>

MTS_PY_EXPORT(Endpoint) {
    using enoki::vectorize_wrapper;

    auto endpoint = MTS_PY_CLASS(Endpoint, Object)
        .def("sample_ray",
             py::overload_cast<Float, const Spectrumf &, const Point2f &,
                               const Point2f &, bool>(&Endpoint::sample_ray, py::const_),
             D(Endpoint, sample_ray),
             "time"_a, "wavelengths"_a, "sample1"_a, "sample2"_a, "unused"_a = true)
        .def("sample_ray",
             enoki::vectorize_wrapper(
                py::overload_cast<FloatP, const SpectrumfP &, const Point2fP &,
                                  const Point2fP &, MaskP>(&Endpoint::sample_ray, py::const_)),
             "time"_a, "wavelengths"_a, "sample1"_a, "sample2"_a, "active"_a = MaskP(true))
#if defined(MTS_ENABLE_AUTODIFF)
        .def("sample_ray",
            py::overload_cast<FloatD, const SpectrumfD &, const Point2fD &, const Point2fD &, MaskD>(
                &Endpoint::sample_ray, py::const_),
             "time"_a, "wavelengths"_a, "sample1"_a, "sample2"_a, "active"_a = true)
#endif

        .def("sample_direction",
             py::overload_cast<const Interaction3f &, const Point2f &, bool>(
                &Endpoint::sample_direction, py::const_),
             D(Endpoint, sample_direction), "it"_a, "sample"_a, "unused"_a = true)
        .def("sample_direction",
             enoki::vectorize_wrapper(
                py::overload_cast<const Interaction3fP &, const Point2fP &, MaskP>(
                    &Endpoint::sample_direction, py::const_)),
             "it"_a, "sample"_a, "active"_a = true)

        .def("sample_position",
             py::overload_cast<Float, const Point2f &, bool>(
                &Endpoint::sample_position, py::const_),
             "time"_a, "sample"_a, "active"_a = true, D(Endpoint, sample_position))
        .def("sample_position",
             enoki::vectorize_wrapper(
                py::overload_cast<FloatP, const Point2fP &, MaskP>(
                    &Endpoint::sample_position, py::const_)
             ), "time"_a, "sample"_a, "active"_a = true,
             D(Endpoint, sample_position, 2))

        .def("sample_wavelengths",
             py::overload_cast<Float, bool>(
                &Endpoint::sample_wavelengths, py::const_),
             D(Endpoint, sample_wavelengths), "sample"_a, "active"_a = true)
        .def("sample_wavelengths",
             enoki::vectorize_wrapper(py::overload_cast<const FloatP &, MaskP>(
                    &Endpoint::sample_wavelengths, py::const_)
             ), "sample"_a, "active"_a = true)
#if defined(MTS_ENABLE_AUTODIFF)
        .def("sample_wavelengths",
             py::overload_cast<const FloatD &, MaskD>(
                    &Endpoint::sample_wavelengths, py::const_
             ), "sample"_a, "active"_a = true)
#endif

        .def("pdf_direction",
             py::overload_cast<const Interaction3f &, const DirectionSample3f &,
                               bool>(&Endpoint::pdf_direction, py::const_),
             D(Endpoint, pdf_direction), "it"_a, "ds"_a, "unused"_a = true)
        .def("pdf_direction",
             enoki::vectorize_wrapper(
                py::overload_cast<const Interaction3fP &,
                                  const DirectionSample3fP &, MaskP>(
                    &Endpoint::pdf_direction, py::const_)
             ), "it"_a, "ds"_a, "active"_a = true)

        .def("pdf_position",
             py::overload_cast<const PositionSample3f &, bool>(
                &Endpoint::pdf_position, py::const_),
             "ps"_a, "active"_a = true, D(Endpoint, pdf_position))
        .def("pdf_position",
             enoki::vectorize_wrapper(py::overload_cast<const PositionSample3fP &, MaskP>(
                    &Endpoint::pdf_position, py::const_)
             ), "ps"_a, "active"_a = true, D(Endpoint, pdf_position))

        .def("pdf_wavelengths",
             py::overload_cast<const Spectrumf &, bool>(
                &Endpoint::pdf_wavelengths, py::const_),
             D(Endpoint, pdf_wavelengths), "wavelengths"_a, "active"_a = true)
        .def("pdf_wavelengths",
             enoki::vectorize_wrapper(py::overload_cast<const SpectrumfP &, MaskP>(
                    &Endpoint::pdf_wavelengths, py::const_)
             ), "wavelengths"_a, "active"_a = true)
#if defined(MTS_ENABLE_AUTODIFF)
        .def("pdf_wavelengths",
             py::overload_cast<const SpectrumfD &, MaskD>(
                &Endpoint::pdf_wavelengths, py::const_),
             D(Endpoint, pdf_wavelengths), "wavelengths"_a, "active"_a = true)
#endif

        .def("eval",
             py::overload_cast<const SurfaceInteraction3f &, bool>(
                &Endpoint::eval, py::const_),
             D(Endpoint, eval), "si"_a, "unused"_a = true)
        .def("eval",
             enoki::vectorize_wrapper(
                py::overload_cast<const SurfaceInteraction3fP &, MaskP>(
                    &Endpoint::eval, py::const_)),
             "si"_a, "active"_a = true)
#if defined(MTS_ENABLE_AUTODIFF)
        .def("eval",
             py::overload_cast<const SurfaceInteraction3fD &, MaskD>(
                &Endpoint::eval, py::const_),
             D(Endpoint, eval), "si"_a, "active"_a = true)
#endif

        .def("sample_ray_pol",
            py::overload_cast<Float, Float, const Point2f &, const Point2f &, bool>(
                &Endpoint::sample_ray_pol, py::const_),
            D(Endpoint, sample_ray_pol),
            "time"_a, "sample1"_a, "sample2"_a, "sample3"_a, "unused"_a = true)
        .def("sample_ray_pol",
            enoki::vectorize_wrapper(
                py::overload_cast<FloatP, FloatP, const Point2fP &, const Point2fP &, MaskP>(
                    &Endpoint::sample_ray_pol, py::const_)),
            "time"_a, "sample1"_a, "sample2"_a, "sample3"_a, "active"_a = true)
        .def("sample_direction_pol",
            py::overload_cast<const Interaction3f &, const Point2f &, bool>(
                &Endpoint::sample_direction_pol, py::const_),
            D(Endpoint, sample_direction_pol), "it"_a, "sample"_a, "unused"_a = true)
        .def("sample_direction_pol",
            enoki::vectorize_wrapper(
                py::overload_cast<const Interaction3fP &, const Point2fP &, MaskP>(
                    &Endpoint::sample_direction_pol, py::const_)),
            "it"_a, "sample"_a, "active"_a = true)
        .def("eval_pol",
            py::overload_cast<const SurfaceInteraction3f &, bool>(
                &Endpoint::eval_pol, py::const_),
            D(Endpoint, eval_pol), "si"_a, "unused"_a = true)
        .def("eval_pol",
            enoki::vectorize_wrapper(
                py::overload_cast<const SurfaceInteraction3fP &, MaskP>(
                    &Endpoint::eval_pol, py::const_)),
            "si"_a, "active"_a = MaskP(true))

        .mdef(Endpoint, world_transform)
        .mdef(Endpoint, needs_sample_1)
        .mdef(Endpoint, needs_sample_2)
        .def("shape", py::overload_cast<>(&Endpoint::shape, py::const_),
             D(Endpoint, shape))
        .def("medium", py::overload_cast<>(&Endpoint::medium, py::const_),
             D(Endpoint, medium))
        .mdef(Endpoint, set_shape, "shape"_a)
        .mdef(Endpoint, set_medium, "medium"_a)
        .mdef(Endpoint, bbox);
}
