#include <mitsuba/python/python.h>
#include <mitsuba/render/medium.h>
#include <mitsuba/render/sampler.h>
#include <mitsuba/render/scene.h>
#include <mitsuba/render/shape.h>

MTS_PY_EXPORT(Medium) {
    auto medium =
        MTS_PY_CLASS(Medium, DifferentiableObject)
            .def("sample_distance",
                 py::overload_cast<const Scene *, const Ray3f &, const Point2f &, Sampler *>(
                     &Medium::sample_distance, py::const_),
                 "scene"_a, "ray"_a, "sample"_a, "sampler"_a, D(Medium, sample_distance))
            .def("sample_distance",
                 enoki::vectorize_wrapper(
                     py::overload_cast<const Scene *, const Ray3fP &, const Point2fP &, Sampler *,
                                       UInt32P, MaskP>(&Medium::sample_distance, py::const_)),
                 "scene"_a, "ray"_a, "sample"_a, "sampler"_a, "index"_a, "active"_a = true,
                 D(Medium, sample_distance))
            .def("eval_transmittance",
                 py::overload_cast<const Ray3f &, Sampler *>(&Medium::eval_transmittance,
                                                             py::const_),
                 "ray"_a, "sampler"_a, D(Medium, eval_transmittance))
            .def("eval_transmittance",
                 enoki::vectorize_wrapper(
                     py::overload_cast<const Ray3fP &, Sampler *, UInt32P, MaskP>(
                         &Medium::eval_transmittance, py::const_)),
                 "ray"_a, "sampler"_a, "index"_a, "active"_a = true, D(Medium, eval_transmittance))
            .def("eval_scattering",
                 py::overload_cast<const MediumInteraction3f &>(&Medium::eval_scattering, py::const_),
                 "mi"_a)
            .def("eval_albedo",
                 py::overload_cast<const Interaction3f &>(&Medium::eval_albedo, py::const_),
                 "it"_a)
            .def("pdf_distance",
                 py::overload_cast<const Ray3f &, float>(&Medium::pdf_distance, py::const_),
                 "ray"_a, "t"_a)
            .def("density",
                 py::overload_cast<const Interaction3f &>(&Medium::density, py::const_),
                 "it"_a)
            .mdef(Medium, is_homogeneous)
            .mdef(Medium, use_emitter_sampling);
}
