#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/properties.h>

NAMESPACE_BEGIN(mitsuba)

#if defined(MTS_ENABLE_AUTODIFF)
using SpectrumfC            = typename ContinuousSpectrum::SpectrumfC;
using MaskC                 = typename ContinuousSpectrum::MaskC;
using SurfaceInteraction3fC = typename ContinuousSpectrum::SurfaceInteraction3fC;
#endif

// =======================================================================
//! @{ \name Spectrum implementation (just throws exceptions)
// =======================================================================


ContinuousSpectrum::ContinuousSpectrum(const Properties &props)
    : m_id(props.id()) {}

ContinuousSpectrum::~ContinuousSpectrum() { }

Spectrumf ContinuousSpectrum::eval(const Spectrumf &) const {
    NotImplementedError("eval");
}

SpectrumfP ContinuousSpectrum::eval(const SpectrumfP &, MaskP) const {
    NotImplementedError("eval_p");
}

#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfC ContinuousSpectrum::eval(const SpectrumfC &, MaskC) const {
    NotImplementedError("eval_c");
}
SpectrumfD ContinuousSpectrum::eval(const SpectrumfD &, MaskD) const {
    NotImplementedError("eval_d");
}
#endif

std::pair<Spectrumf, Spectrumf>
ContinuousSpectrum::sample(const Spectrumf &) const {
    NotImplementedError("sample");
}

std::pair<SpectrumfP, SpectrumfP> ContinuousSpectrum::sample(const SpectrumfP &,
                                                             MaskP) const {
    NotImplementedError("sample_p");
}

#if defined(MTS_ENABLE_AUTODIFF)
std::pair<SpectrumfC, SpectrumfC> ContinuousSpectrum::sample(const SpectrumfC &,
                                                             MaskC) const {
    NotImplementedError("sample_c");
}
std::pair<SpectrumfD, SpectrumfD> ContinuousSpectrum::sample(const SpectrumfD &,
                                                             MaskD) const {
    NotImplementedError("sample_d");
}
#endif

Spectrumf ContinuousSpectrum::pdf(const Spectrumf &) const {
    NotImplementedError("pdf");
}

SpectrumfP ContinuousSpectrum::pdf(const SpectrumfP &, MaskP) const {
    NotImplementedError("pdf_p");
}


#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfC ContinuousSpectrum::pdf(const SpectrumfC &, MaskC) const {
    NotImplementedError("pdf_c");
}
SpectrumfD ContinuousSpectrum::pdf(const SpectrumfD &, MaskD) const {
    NotImplementedError("pdf_d");
}
#endif

Float ContinuousSpectrum::mean() const {
    NotImplementedError("mean");
}

Float ContinuousSpectrum::max() const {
    NotImplementedError("max");
}

Spectrumf ContinuousSpectrum::eval(const SurfaceInteraction3f &si) const {
    return eval(si.wavelengths);
}

SpectrumfP ContinuousSpectrum::eval(const SurfaceInteraction3fP &si, MaskP active) const {
    return eval(si.wavelengths, active);
}

#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfC ContinuousSpectrum::eval(const SurfaceInteraction3fC &si, MaskC active) const {
    return eval(si.wavelengths, active);
}
SpectrumfD ContinuousSpectrum::eval(const SurfaceInteraction3fD &si, MaskD active) const {
    return eval(si.wavelengths, active);
}
#endif

std::pair<Spectrumf, Spectrumf>
ContinuousSpectrum::sample(const SurfaceInteraction3f &, const Spectrumf &sample_) const {
    return sample(sample_);
}

std::pair<SpectrumfP, SpectrumfP>
ContinuousSpectrum::sample(const SurfaceInteraction3fP &, const SpectrumfP &sample_, MaskP active) const {
    return sample(sample_, active);
}

#if defined(MTS_ENABLE_AUTODIFF)
std::pair<SpectrumfC, SpectrumfC>
ContinuousSpectrum::sample(const SurfaceInteraction3fC &, const SpectrumfC &sample_, MaskC active) const {
    return sample(sample_, active);
}
std::pair<SpectrumfD, SpectrumfD>
ContinuousSpectrum::sample(const SurfaceInteraction3fD &, const SpectrumfD &sample_, MaskD active) const {
    return sample(sample_, active);
}
#endif

Spectrumf ContinuousSpectrum::pdf(const SurfaceInteraction3f &si) const {
    return pdf(si.wavelengths);
}

SpectrumfP ContinuousSpectrum::pdf(const SurfaceInteraction3fP &si, MaskP active) const {
    return pdf(si.wavelengths, active);
}

#if defined(MTS_ENABLE_AUTODIFF)
SpectrumfC ContinuousSpectrum::pdf(const SurfaceInteraction3fC &si, MaskC active) const {
    return pdf(si.wavelengths, active);
}
SpectrumfD ContinuousSpectrum::pdf(const SurfaceInteraction3fD &si, MaskD active) const {
    return pdf(si.wavelengths, active);
}
#endif

ref<ContinuousSpectrum> ContinuousSpectrum::D65(Float scale, EColorMode color_mode) {
    if (unlikely(color_mode != ESpectral)) {
        Properties props("uniform");
        props.set_int("color_mode", color_mode);
        props.mark_queried("color_mode");
        // Should output 1 by default.
        Float correction =
            (color_mode == EMonochrome) ? 1.f / (MTS_WAVELENGTH_MAX - MTS_WAVELENGTH_MIN) : 1.f;
        props.set_float("value", scale * correction);
        auto obj = PluginManager::instance()->create_object<ContinuousSpectrum>(props);
        return (ContinuousSpectrum *) (obj.get());
    }

    Properties props("d65");
    props.set_float("value", scale);
    auto obj =
        PluginManager::instance()->create_object<ContinuousSpectrum>(props);
    return (ContinuousSpectrum *) (obj->expand()[0].get());
}

std::string ContinuousSpectrum::id() const {
    return m_id;
}

//! @}
// =======================================================================

MTS_IMPLEMENT_CLASS_ALIAS(ContinuousSpectrum, "spectrum", Object)
NAMESPACE_END(mitsuba)
