import numpy as np

import mitsuba
from mitsuba.core.xml import load_string


def test01_create():
    # p = load_string("<medium version='2.0.0' type='heterogeneous'>")
    # assert p is not None

    p = load_string("""<medium version='2.0.0' type='heterogeneous'>
        <texture3d name="density" type="constant3d">
            <spectrum name="color" value="2.5"/>
        </texture3d>
    </medium>""")
    assert p is not None


def test02_pdf():
    from mitsuba.core import Ray3f, BoundingBox3f, EColorMode

    # Uses the unit bounding box by default: (0, 0, 0) -> (1, 1, 1)
    medium = load_string("""<medium version='2.0.0' type='heterogeneous'>
        <texture3d name="density" type="constant3d">
            <spectrum name="color" value="3"/>
        </texture3d>
        <boolean name="use_raymarching" value="true"/>
    </medium>""", color_mode=EColorMode.ERGBColor)

    ray = Ray3f()
    ray.o = [0.5, 0.5, 0.5]
    ray.d = [0., 0., 1.]
    ray.d = ray.d / np.linalg.norm(ray.d)
    ray.update()
    t = 0.25
    print(ray)

    tr_ray = Ray3f(ray)
    tr_ray.mint = 0
    tr_ray.maxt = t
    bbox = BoundingBox3f([0, 0, 0], [1, 1, 1])
    print(bbox)
    has_it, aabb_mint, aabb_maxt = bbox.ray_intersect(tr_ray)
    print(f'has_it: {has_it}, aabb_mint: {aabb_mint}, aabb_maxt: {aabb_maxt}')

    expected = np.exp(-3 * t)
    wrong = np.exp(-3 * 0.5)
    found = medium.pdf_distance(ray, t)

    print(f'expected: {expected}, found: {found}, wrong: {wrong}')
    assert np.allclose(expected, found)
