import numpy as np

import mitsuba
from mitsuba.core.math import Pi
from mitsuba.core.xml import load_string
from mitsuba.core import MTS_WAVELENGTH_SAMPLES, Ray3f

def test01_create():
    p = load_string("<medium version='2.0.0' type='homogeneous'/>")
    assert p is not None


def test02_eval_tr():
    sigma_t = 0.5
    density = 4
    p = load_string("""<medium version='2.0.0' type='homogeneous'>
                        <phase type='isotropic' />
                        <spectrum name='sigma_t' type='uniform'>
                            <float name="value" value="{0}"/>
                        </spectrum>
                        <float name='density' value='{1}'/>
                    </medium>""".format(sigma_t, density))
    ray = Ray3f()
    ray.wavelengths = np.ones(MTS_WAVELENGTH_SAMPLES) * 500
    sampler = load_string("<sampler version='2.0.0' type='independent'/>")
    for l in [0.0, 0.5, 5, 10]:
        ray.mint = 0
        ray.maxt = l
        eval_tr = p.eval_transmittance(ray, sampler)
        eval_tr = np.mean(eval_tr)
        ref_tr = np.exp(-(ray.maxt - ray.mint) * sigma_t * density)
        assert np.allclose(eval_tr, ref_tr)



# def test03_sample_dist():
#     sigma_t = 0.5
#     density = 4
#     albedo = 0.8

#     # Create a a scene with a box, filled with medium
#     # Then: call sample_distance

#     scene = load_string(f"""
#         <scene version="2.0.0">
#             <shape type="ply">
#                 <string name="filename" value="resources/data/ply/cube.ply"/>
#                     <medium name='interior' type='homogeneous'>
#                         <phase type='isotropic' />
#                         <spectrum name='sigma_t' type='uniform'>
#                             <float name="value" value="{sigma_t}"/>
#                         </spectrum>
#                         <spectrum name='albedo' type='uniform'>
#                             <float name="value" value="{albedo}"/>
#                         </spectrum>
#                         <float name='density' value='{density}'/>
#                     </medium>
#                 </shape>
#         </scene>
#     """)
#     sampler = load_string("<sampler version='2.0.0' type='independent'/>")
#     shapes = scene.shapes()
#     shape = shapes[0]
#     medium = shape.interior_medium()

#     ray = Ray3f()
#     ray.o = np.array([0,0,0])
#     ray.d = np.array([0,1,0])
#     ray.time = 0
#     print(dir(mitsuba.render))

#     ray.wavelengths = np.ones(MTS_WAVELENGTH_SAMPLES) * 500

#     ret = medium.sample_distance(scene, ray, sampler)
#     si, mi = ret[0]
#     throughput = ret[1]
#     print(throughput)
#     assert False
