/**
 * @file Wrappers for calls to PTX-extracted BSDF methods.
 */

#pragma once

#include "types.cuh"

namespace detail {
extern "C" {
__device__ void bsdf_sample(
    float,     // si.t
    float,     // si.time
    float,     // si.wavelengths.0
    float,     // si.wavelengths.1
    float,     // si.wavelengths.2
    float,     // si.wavelengths.3
    float,     // si.p.0
    float,     // si.p.1
    float,     // si.p.2
    void *,    // si.shape
    float,     // si.uv.0
    float,     // si.uv.1
    float,     // si.n.0
    float,     // si.n.1
    float,     // si.n.2
    float,     // si.sh_frame.s.0
    float,     // si.sh_frame.s.1
    float,     // si.sh_frame.s.2
    float,     // si.sh_frame.t.0
    float,     // si.sh_frame.t.1
    float,     // si.sh_frame.t.2
    float,     // si.sh_frame.n.0
    float,     // si.sh_frame.n.1
    float,     // si.sh_frame.n.2
    float,     // si.dp_du.0
    float,     // si.dp_du.1
    float,     // si.dp_du.2
    float,     // si.dp_dv.0
    float,     // si.dp_dv.1
    float,     // si.dp_dv.2
    float,     // si.duv_dx.0
    float,     // si.duv_dx.1
    float,     // si.duv_dy.0
    float,     // si.duv_dy.1
    float,     // si.wi.0
    float,     // si.wi.1
    float,     // si.wi.2
    uint32_t,  // si.prim_index
    uint64_t,  // si.instance
    float,     // sample1
    float,     // sample2.0
    float,     // sample2.1
    float *,   // bs.wo.0
    float *,   // bs.wo.1
    float *,   // bs.wo.2
    float *,   // bs.pdf
    float *,   // bs.eta
    uint32_t *,// bs.sampled_type
    uint32_t *,// bs.sampled_component
    float *,   // weight.0
    float *,   // weight.1
    float *,   // weight.2
    float *    // weight.3
);

__device__ void bsdf_eval(
    float,    // si.t
    float,    // si.time
    float,    // si.wavelengths.0
    float,    // si.wavelengths.1
    float,    // si.wavelengths.2
    float,    // si.wavelengths.3
    float,    // si.p.0
    float,    // si.p.1
    float,    // si.p.2
    void *,   // si.shape
    float,    // si.uv.0
    float,    // si.uv.1
    float,    // si.n.0
    float,    // si.n.1
    float,    // si.n.2
    float,    // si.sh_frame.s.0
    float,    // si.sh_frame.s.1
    float,    // si.sh_frame.s.2
    float,    // si.sh_frame.t.0
    float,    // si.sh_frame.t.1
    float,    // si.sh_frame.t.2
    float,    // si.sh_frame.n.0
    float,    // si.sh_frame.n.1
    float,    // si.sh_frame.n.2
    float,    // si.dp_du.0
    float,    // si.dp_du.1
    float,    // si.dp_du.2
    float,    // si.dp_dv.0
    float,    // si.dp_dv.1
    float,    // si.dp_dv.2
    float,    // si.duv_dx.0
    float,    // si.duv_dx.1
    float,    // si.duv_dy.0
    float,    // si.duv_dy.1
    float,    // si.wi.0
    float,    // si.wi.1
    float,    // si.wi.2
    uint32_t, // si.prim_index
    uint64_t, // si.instance
    float,    // wo.0
    float,    // wo.1
    float,    // wo.2
    float *,  // result.0
    float *,  // result.1
    float *,  // result.2
    float *   // result.3
);

__device__ void bsdf_pdf(
    float,      // si.t
    float,      // si.time
    float,      // si.wavelengths.0
    float,      // si.wavelengths.1
    float,      // si.wavelengths.2
    float,      // si.wavelengths.3
    float,      // si.p.0
    float,      // si.p.1
    float,      // si.p.2
    void *,     // si.shape
    float,      // si.uv.0
    float,      // si.uv.1
    float,      // si.n.0
    float,      // si.n.1
    float,      // si.n.2
    float,      // si.sh_frame.s.0
    float,      // si.sh_frame.s.1
    float,      // si.sh_frame.s.2
    float,      // si.sh_frame.t.0
    float,      // si.sh_frame.t.1
    float,      // si.sh_frame.t.2
    float,      // si.sh_frame.n.0
    float,      // si.sh_frame.n.1
    float,      // si.sh_frame.n.2
    float,      // si.dp_du.0
    float,      // si.dp_du.1
    float,      // si.dp_du.2
    float,      // si.dp_dv.0
    float,      // si.dp_dv.1
    float,      // si.dp_dv.2
    float,      // si.duv_dx.0
    float,      // si.duv_dx.1
    float,      // si.duv_dy.0
    float,      // si.duv_dy.1
    float,      // si.wi.0
    float,      // si.wi.1
    float,      // si.wi.2
    uint32_t,   // si.prim_index
    uint64_t,   // si.instance
    float,      // wo.0
    float,      // wo.1
    float,      // wo.2
    float *     // result2
);

}
}  // namespace detail


struct BSDFSample {
    optix::float3 sampling_weight;
    optix::float3 wo;
    float pdf;
    float eta;
    uint32_t sampled_type;
    uint32_t sampled_component;
};

RT_FUNCTION BSDFSample call_bsdf_sample(const SurfaceInteraction &si,
                                        float sample1, const optix::float2 &sample2) {
    BSDFSample bs;
    optix::float4 weight4;
    detail::bsdf_sample(
        0.f,                           // si.t
        0.f,                           // si.time
        0.f, 0.f, 0.f, 0.f,            // si.wavelengths.0
        si.p.x, si.p.y, si.p.z,
        nullptr,                       // si.shape
        si.uv.x, si.uv.y,
        si.normal.x, si.normal.y, si.normal.z,
        si.sh_frame.s.x, si.sh_frame.s.y, si.sh_frame.s.z,
        si.sh_frame.t.x, si.sh_frame.t.y, si.sh_frame.t.z,
        si.sh_frame.n.x, si.sh_frame.n.y, si.sh_frame.n.z,
        0.f, 0.f, 0.f,                 // si.dp_du
        0.f, 0.f, 0.f,                 // si.dp_dv
        0.f, 0.f,                      // si.duv_dx
        0.f, 0.f,                      // si.duv_dy
        si.wi.x, si.wi.y, si.wi.z,     // si.wi
        0,                             // si.prim_index
        0,                             // si.instance
        sample1,
        sample2.x, sample2.y,
        // Outputs
        &bs.wo.x, &bs.wo.y, &bs.wo.z,
        &bs.pdf, &bs.eta, &bs.sampled_type, &bs.sampled_component,
        &weight4.x, &weight4.y, &weight4.z, &weight4.w
    );
    bs.sampling_weight = optix::make_float3(weight4.x, weight4.y, weight4.z);
    return bs;
}

RT_FUNCTION optix::float3 call_bsdf_eval(const SurfaceInteraction &si,
                                         const optix::float3 &wo) {
    optix::float4 value4;
    detail::bsdf_eval(
        0.f,                           // si.t
        0.f,                           // si.time
        0.f, 0.f, 0.f, 0.f,            // si.wavelengths.0
        si.p.x, si.p.y, si.p.z,
        nullptr,                       // si.shape
        si.uv.x, si.uv.y,
        si.normal.x, si.normal.y, si.normal.z,
        si.sh_frame.s.x, si.sh_frame.s.y, si.sh_frame.s.z,
        si.sh_frame.t.x, si.sh_frame.t.y, si.sh_frame.t.z,
        si.sh_frame.n.x, si.sh_frame.n.y, si.sh_frame.n.z,
        0.f, 0.f, 0.f,                 // si.dp_du
        0.f, 0.f, 0.f,                 // si.dp_dv
        0.f, 0.f,                      // si.duv_dx
        0.f, 0.f,                      // si.duv_dy
        si.wi.x, si.wi.y, si.wi.z,     // si.wi
        0,                             // si.prim_index
        0,                             // si.instance
        wo.x, wo.y, wo.z,
        // Outputs
        &value4.x, &value4.y, &value4.z, &value4.w
    );

    return optix::make_float3(value4.x, value4.y, value4.z);
}

RT_FUNCTION float call_bsdf_pdf(const SurfaceInteraction &si, const optix::float3 &wo) {
    float pdf;
    detail::bsdf_pdf(
        0.f,                           // si.t
        0.f,                           // si.time
        0.f, 0.f, 0.f, 0.f,            // si.wavelengths.0
        si.p.x, si.p.y, si.p.z,
        nullptr,                       // si.shape
        si.uv.x, si.uv.y,
        si.normal.x, si.normal.y, si.normal.z,
        si.sh_frame.s.x, si.sh_frame.s.y, si.sh_frame.s.z,
        si.sh_frame.t.x, si.sh_frame.t.y, si.sh_frame.t.z,
        si.sh_frame.n.x, si.sh_frame.n.y, si.sh_frame.n.z,
        0.f, 0.f, 0.f,                 // si.dp_du
        0.f, 0.f, 0.f,                 // si.dp_dv
        0.f, 0.f,                      // si.duv_dx
        0.f, 0.f,                      // si.duv_dy
        si.wi.x, si.wi.y, si.wi.z,     // si.wi
        0,                             // si.prim_index
        0,                             // si.instance
        wo.x, wo.y, wo.z,
        // Outputs
        &pdf
    );
    return pdf;
}
