/**
 * @file Wrappers for calls to PTX-extracted BSDF autodiff methods.
 */

#pragma once

#include "types.cuh"

namespace detail {
extern "C" {
__device__ void bsdf_eval_d(
    float,    // si.t
    float,    // si.time
    float,    // si.wavelengths.0
    float,    // si.wavelengths.1
    float,    // si.wavelengths.2
    float,    // si.wavelengths.3
    float,    // si.p.0
    float,    // si.p.1
    float,    // si.p.2
    void *,   // si.shape
    float,    // si.uv.0
    float,    // si.uv.1
    float,    // si.n.0
    float,    // si.n.1
    float,    // si.n.2
    float,    // si.sh_frame.s.0
    float,    // si.sh_frame.s.1
    float,    // si.sh_frame.s.2
    float,    // si.sh_frame.t.0
    float,    // si.sh_frame.t.1
    float,    // si.sh_frame.t.2
    float,    // si.sh_frame.n.0
    float,    // si.sh_frame.n.1
    float,    // si.sh_frame.n.2
    float,    // si.dp_du.0
    float,    // si.dp_du.1
    float,    // si.dp_du.2
    float,    // si.dp_dv.0
    float,    // si.dp_dv.1
    float,    // si.dp_dv.2
    float,    // si.duv_dx.0
    float,    // si.duv_dx.1
    float,    // si.duv_dy.0
    float,    // si.duv_dy.1
    float,    // si.wi.0
    float,    // si.wi.1
    float,    // si.wi.2
    uint32_t, // si.prim_index
    uint64_t, // si.instance
    float,    // wo.0
    float,    // wo.1
    float,    // wo.2
    float,    // grad_of_result.0
    float,    // grad_of_result.1
    float,    // grad_of_result.2
    float     // grad_of_result.3
    );

}
}  // namespace detail


RT_FUNCTION void call_bsdf_eval_d(const SurfaceInteraction &si, const optix::float3 &wo,
                                  const optix::float3 &grad_of_result) {
    detail::bsdf_eval_d(
        0.f,                           // si.t
        0.f,                           // si.time
        0.f, 0.f, 0.f, 0.f,            // si.wavelengths.0
        si.p.x, si.p.y, si.p.z,
        nullptr,                       // si.shape
        si.uv.x, si.uv.y,
        si.normal.x, si.normal.y, si.normal.z,
        si.sh_frame.s.x, si.sh_frame.s.y, si.sh_frame.s.z,
        si.sh_frame.t.x, si.sh_frame.t.y, si.sh_frame.t.z,
        si.sh_frame.n.x, si.sh_frame.n.y, si.sh_frame.n.z,
        0.f, 0.f, 0.f,                 // si.dp_du
        0.f, 0.f, 0.f,                 // si.dp_dv
        0.f, 0.f,                      // si.duv_dx
        0.f, 0.f,                      // si.duv_dy
        si.wi.x, si.wi.y, si.wi.z,     // si.wi
        0,                             // si.prim_index
        0,                             // si.instance
        wo.x, wo.y, wo.z,
        grad_of_result.x, grad_of_result.y, grad_of_result.z, 0.f
        );
}
