#pragma once

#include <limits>
#include <optix_world.h>
#include <math_constants.h>
#include "random.cuh"

namespace math {
constexpr float  Epsilon         = 1e-4f;
constexpr float  ShadowEpsilon   = 1e-3f;
constexpr float  Pi              = (float) 3.14159265358979323846;
constexpr float  InvPi           = (float) 0.31830988618379067154;
constexpr float  Infinity        = std::numeric_limits<float>::infinity();
}  // end namespace math

__device__ inline float squared_norm(optix::float3 v) {
    return optix::dot(v, v);
}

__device__ __inline__ float power_heuristic(float a, float b) {
    a *= a;
    b *= b;

    if (a == 0.f) return 0.f;

    return a / (a + b);
}

/** Creates a new standard ray starting off from a surface. */
__device__ __inline__ optix::Ray spawn_ray(const optix::float3 &origin,
                                           const optix::float3 &direction) {
    float max_origin = optix::fmaxf(optix::fmaxf(abs(origin.x), abs(origin.y)), abs(origin.z));
    float min_t = math::Epsilon * (1.f + max_origin);
    // Ray type 1: standard ray
    return optix::make_Ray(origin, direction, 1, min_t, RT_DEFAULT_MAX);
}

/** Creates a new shadow ray starting off from a surface. */
__device__ __inline__ optix::Ray spawn_shadow_ray(const optix::float3 &origin,
                                                  const optix::float3 &direction, float max_dist) {
    float max_origin = optix::fmaxf(optix::fmaxf(abs(origin.x), abs(origin.y)), abs(origin.z));
    float min_t = math::Epsilon * (1.f + max_origin);
    float max_t = max_dist * (1.f - math::ShadowEpsilon);
    // Ray type 2: shadow ray
    return optix::make_Ray(origin, direction, 2, min_t, max_t);
}
