/**
* @file Wrappers for calls to PTX-extracted Envmap methods.
*/

#pragma once

#include <optix_world.h>
// #include "types.cuh"

namespace detail {
extern "C" {
__device__ void envmap_eval(
    float,        // si.t
    float,        // si.time
    float,        // si.wavelengths.0
    float,        // si.wavelengths.1
    float,        // si.wavelengths.2
    float,        // si.wavelengths.3
    float,        // si.p.0
    float,        // si.p.1
    float,        // si.p.2
    void *,       // si.shape
    float,        // si.uv.0
    float,        // si.uv.1
    float,        // si.n.0
    float,        // si.n.1
    float,        // si.n.2
    float,        // si.sh_frame.s.0
    float,        // si.sh_frame.s.1
    float,        // si.sh_frame.s.2
    float,        // si.sh_frame.t.0
    float,        // si.sh_frame.t.1
    float,        // si.sh_frame.t.2
    float,        // si.sh_frame.n.0
    float,        // si.sh_frame.n.1
    float,        // si.sh_frame.n.2
    float,        // si.dp_du.0
    float,        // si.dp_du.1
    float,        // si.dp_du.2
    float,        // si.dp_dv.0
    float,        // si.dp_dv.1
    float,        // si.dp_dv.2
    float,        // si.duv_dx.0
    float,        // si.duv_dx.1
    float,        // si.duv_dy.0
    float,        // si.duv_dy.1
    float,        // si.wi.0
    float,        // si.wi.1
    float,        // si.wi.2
    uint32_t,     // si.prim_index
    void *,       // si.instance
    float *,      // result.0
    float *,      // result.1
    float *,      // result.2
    float *	      // result.3
);

__device__ void envmap_sample_direction(
    float,    // ref.t
    float,    // ref.time
    float,    // ref.wavelengths.0
    float,    // ref.wavelengths.1
    float,    // ref.wavelengths.2
    float,    // ref.wavelengths.3
    float,    // ref.p.0
    float,    // ref.p.1
    float,    // ref.p.2
    float,    // sample.0
    float,    // sample.1
    float *,  // ds.p.0
    float *,  // ds.p.1
    float *,  // ds.p.2
    float *,  // ds.n.0
    float *,  // ds.n.1
    float *,  // ds.n.2
    float *,  // ds.uv.0
    float *,  // ds.uv.1
    float *,  // ds.time
    float *,  // ds.pdf
    bool *,   // ds.delta
    void **,  // ds.object
    float *,  // ds.d.0
    float *,  // ds.d.1
    float *,  // ds.d.2
    float *,  // ds.dist
    float *,  // result.0
    float *,  // result.1
    float *,  // result.2
    float *	  // result.3
);

__device__ void envmap_pdf_direction(
    float,  // it.t
    float,  // it.time
    float,  // it.wavelengths.0
    float,  // it.wavelengths.1
    float,  // it.wavelengths.2
    float,  // it.wavelengths.3
    float,  // it.p.0
    float,  // it.p.1
    float,  // it.p.2
    float,  // ds.p.0
    float,  // ds.p.1
    float,  // ds.p.2
    float,  // ds.n.0
    float,  // ds.n.1
    float,  // ds.n.2
    float,  // ds.uv.0
    float,  // ds.uv.1
    float,  // ds.time
    float,  // ds.pdf
    float,  // ds.delta (note: workaround applied to support bool types, `delta` unused anyway)
    void*,  // ds.object
    float,  // ds.d.0
    float,  // ds.d.1
    float,  // ds.d.2
    float,  // ds.dist
    float *   // result
);

}  // end extern
}  // end namespace detail


/// Takes a direction pointing toward the envmap, in world coordinates
RT_FUNCTION optix::float3 call_envmap_eval(const optix::float3 &direction) {
    float f = 0.f;  // For all unused fields

    optix::float4 res4;
    detail::envmap_eval(
        0.f,                           // si.t
        0.f,                           // si.time
        0.f, 0.f, 0.f, 0.f,            // si.wavelengths
        f, f, f,                       // si.p
        nullptr,                       // si.shape
        f, f,                          // si.uv
        f, f, f,                       // si.normal
        f, f, f,                       // si.sh_frame.s
        f, f, f,                       // si.sh_frame.t
        f, f, f,                       // si.sh_frame.n
        0.f, 0.f, 0.f,                 // si.dp_du
        0.f, 0.f, 0.f,                 // si.dp_dv
        0.f, 0.f,                      // si.duv_dx
        0.f, 0.f,                      // si.duv_dy
        -direction.x, -direction.y, -direction.z,  // si.wi
        0,                             // si.prim_index
        nullptr,                       // si.instance
        &res4.x, &res4.y, &res4.z, &res4.w
    );
    // res4 = make_float4(0.f, 0.5f, 0.5f, 0.f);
    return optix::make_float3(res4.x, res4.y, res4.z);
}

/// Samples a direction pointing toward the envmap, in world coordinates
RT_FUNCTION DirectionSample call_envmap_sample_direction(const MediumSample &mi,
                                                         const optix::float2 &sample) {
    float f = 0.f;  // For all unused fields
    bool bl;
    void *ptr = nullptr;

    DirectionSample ds;
    detail::envmap_sample_direction(
        mi.t,
        0.f,                // ref.time
        0.f, 0.f, 0.f, 0.f, // ref.wavelengths
        mi.p.x, mi.p.y, mi.p.z,
        sample.x, sample.y,

        // Outputs
        &ds.p.x, &ds.p.y, &ds.p.z,
        &ds.normal.x, &ds.normal.y, &ds.normal.z,
        &f, &f,   // ds.uv
        &f,       // ds.time
        &ds.pdf,
        &bl,     // ds.delta
        &ptr,    // ds.object
        &ds.light_direction.x, &ds.light_direction.y, &ds.light_direction.z,
        &ds.light_dist,
        &ds.radiance.x, &ds.radiance.y, &ds.radiance.z, &f
    );
    return ds;
}

/// Samples a direction pointing toward the envmap from a surface, in world coordinates
RT_FUNCTION DirectionSample call_envmap_sample_direction(const SurfaceInteraction &si,
                                                         const optix::float2 &sample) {
    float f = 0.f;  // For all unused fields
    bool bl;
    void *ptr = nullptr;

    DirectionSample ds;
    detail::envmap_sample_direction(
        si.t,
        0.f,                // ref.time
        0.f, 0.f, 0.f, 0.f, // ref.wavelengths
        si.p.x, si.p.y, si.p.z,
        sample.x, sample.y,

        // Outputs
        &ds.p.x, &ds.p.y, &ds.p.z,
        &ds.normal.x, &ds.normal.y, &ds.normal.z,
        &f, &f,   // ds.uv
        &f,       // ds.time
        &ds.pdf,
        &bl,     // ds.delta
        &ptr,    // ds.object
        &ds.light_direction.x, &ds.light_direction.y, &ds.light_direction.z,
        &ds.light_dist,
        &ds.radiance.x, &ds.radiance.y, &ds.radiance.z, &f
    );
    return ds;
}

/// Takes a direction pointing toward the envmap, in world coordinates
RT_FUNCTION float call_envmap_pdf_direction(const optix::float3 &direction) {
    float res;
    detail::envmap_pdf_direction(
        // Inputs: it
        0.f,
        0.f,                // ref.time
        0.f, 0.f, 0.f, 0.f, // ref.wavelengths
        0.f, 0.f, 0.f,
        // Inputs: ds
        0.f, 0.f, 0.f,
        0.f, 0.f, 0.f,
        0.f, 0.f,   // ds.uv
        0.f,       // ds.time
        0.f,
        0,     // ds.delta
        nullptr,    // ds.object
        direction.x, direction.y, direction.z,
        0.f,
        // Outputs
        &res
    );
    return res;
}
