/**
 * @file Wrappers for calls to PTX-extracted Media methods.
 */

#pragma once

#include <optix_world.h>
#include "types.cuh"

namespace detail {
extern "C" {
__device__ void medium_eval_transmittance(
    float,     // ray.o.0
    float,     // ray.o.1
    float,     // ray.o.2
    float,     // ray.d.0
    float,     // ray.d.1
    float,     // ray.d.2
    float,     // ray.d_rcp.0
    float,     // ray.d_rcp.1
    float,     // ray.d_rcp.2
    float,     // ray.mint
    float,     // ray.maxt
    float,     // ray.time
    float,     // ray.wavelengths.0
    float,     // ray.wavelengths.1
    float,     // ray.wavelengths.2
    float,     // ray.wavelengths.3
    uint32_t,  // index
    float *,   // result.0
    float *,   // result.1
    float *,   // result.2
    float *    // result.3
);

__device__ void medium_sample_distance(
    float,        // si.t
    float,        // si.time
    float,        // si.wavelengths.0
    float,        // si.wavelengths.1
    float,        // si.wavelengths.2
    float,        // si.wavelengths.3
    float,        // si.p.0
    float,        // si.p.1
    float,        // si.p.2
    void *,       // si.shape
    float,        // si.uv.0
    float,        // si.uv.1
    float,        // si.n.0
    float,        // si.n.1
    float,        // si.n.2
    float,        // si.sh_frame.s.0
    float,        // si.sh_frame.s.1
    float,        // si.sh_frame.s.2
    float,        // si.sh_frame.t.0
    float,        // si.sh_frame.t.1
    float,        // si.sh_frame.t.2
    float,        // si.sh_frame.n.0
    float,        // si.sh_frame.n.1
    float,        // si.sh_frame.n.2
    float,        // si.dp_du.0
    float,        // si.dp_du.1
    float,        // si.dp_du.2
    float,        // si.dp_dv.0
    float,        // si.dp_dv.1
    float,        // si.dp_dv.2
    float,        // si.duv_dx.0
    float,        // si.duv_dx.1
    float,        // si.duv_dy.0
    float,        // si.duv_dy.1
    float,        // si.wi.0
    float,        // si.wi.1
    float,        // si.wi.2
    uint32_t,     // si.prim_index
    void *,       // si.instance
    float,        // ray.o.0
    float,        // ray.o.1
    float,        // ray.o.2
    float,        // ray.d.0
    float,        // ray.d.1
    float,        // ray.d.2
    float,        // ray.d_rcp.0
    float,        // ray.d_rcp.1
    float,        // ray.d_rcp.2
    float,        // ray.mint
    float,        // ray.maxt
    float,        // ray.time
    float,        // ray.wavelengths.0
    float,        // ray.wavelengths.1
    float,        // ray.wavelengths.2
    float,        // ray.wavelengths.3
    float,        // sample.0
    float,        // sample.1
    uint32_t,     // index
    float *,      // mi.t
    float *,      // mi.time
    float *,      // mi.wavelengths.0
    float *,      // mi.wavelengths.1
    float *,      // mi.wavelengths.2
    float *,      // mi.wavelengths.3
    float *,      // mi.p.0
    float *,      // mi.p.1
    float *,      // mi.p.2
    void **,      // mi.medium
    float *,      // mi.sh_frame.s.0
    float *,      // mi.sh_frame.s.1
    float *,      // mi.sh_frame.s.2
    float *,      // mi.sh_frame.t.0
    float *,      // mi.sh_frame.t.1
    float *,      // mi.sh_frame.t.2
    float *,      // mi.sh_frame.n.0
    float *,      // mi.sh_frame.n.1
    float *,      // mi.sh_frame.n.2
    float *,      // mi.wi.0
    float *,      // mi.wi.1
    float *,      // mi.wi.2
    float *,      // throughput.0
    float *,      // throughput.1
    float *,      // throughput.2
    float *	      // throughput.3
);

__device__ void medium_pdf_distance(
    float,     // ray.o.0
    float,     // ray.o.1
    float,     // ray.o.2
    float,     // ray.d.0
    float,     // ray.d.1
    float,     // ray.d.2
    float,     // ray.d_rcp.0
    float,     // ray.d_rcp.1
    float,     // ray.d_rcp.2
    float,     // ray.mint
    float,     // ray.maxt
    float,     // ray.time
    float,     // ray.wavelengths.0
    float,     // ray.wavelengths.1
    float,     // ray.wavelengths.2
    float,     // ray.wavelengths.3
    float,     // t
    float *    // result
);

__device__ void medium_eval_scattering(
    float,        // mi.t
    float,        // mi.time
    float,        // mi.wavelengths.0
    float,        // mi.wavelengths.1
    float,        // mi.wavelengths.2
    float,        // mi.wavelengths.3
    float,        // mi.p.0
    float,        // mi.p.1
    float,        // mi.p.2
    void *,       // mi.medium
    float,        // mi.sh_frame.s.0
    float,        // mi.sh_frame.s.1
    float,        // mi.sh_frame.s.2
    float,        // mi.sh_frame.t.0
    float,        // mi.sh_frame.t.1
    float,        // mi.sh_frame.t.2
    float,        // mi.sh_frame.n.0
    float,        // mi.sh_frame.n.1
    float,        // mi.sh_frame.n.2
    float,        // mi.wi.0
    float,        // mi.wi.1
    float,        // mi.wi.2
    float *,      // result.0
    float *,      // result.1
    float *,      // result.2
    float *       // result.3
);

__device__ void medium_eval_albedo(
    float,        // it.t
    float,        // it.time
    float,        // it.wavelengths.0
    float,        // it.wavelengths.1
    float,        // it.wavelengths.2
    float,        // it.wavelengths.3
    float,        // it.p.0
    float,        // it.p.1
    float,        // it.p.2
    float *,      // result.0
    float *,      // result.1
    float *,      // result.2
    float *       // result.3
);

__device__ void medium_density(
    float,        // it.t
    float,        // it.time
    float,        // it.wavelengths.0
    float,        // it.wavelengths.1
    float,        // it.wavelengths.2
    float,        // it.wavelengths.3
    float,        // it.p.0
    float,        // it.p.1
    float,        // it.p.2
    float *       // result
);

// This function is useful when re-implementing heterogeneous media in Cuda
__device__ void math_solve_quadratic(
    float,        // a
    float,        // b
    float,        // c
    bool *,       // has_solution
    float *,      // solution1
    float *       // solution2
);

}  // end extern
}  // namespace detail


RT_FUNCTION optix::float3 call_medium_eval_transmittance(const optix::Ray &ray) {
    optix::float4 res4;
    detail::medium_eval_transmittance(
        ray.origin.x, ray.origin.y, ray.origin.z,
        ray.direction.x, ray.direction.y, ray.direction.z,
        // ray.d_rcp
        1.f / ray.direction.x, 1.f / ray.direction.y, 1.f / ray.direction.z,
        ray.tmin, ray.tmax,
        // ray.time
        0.f,
        // ray.wavelengths
        0.f, 0.f, 0.f, 0.f,
        // index (assumed unused!)
        0,
        // Outputs
        &res4.x, &res4.y, &res4.z, &res4.w
    );
    return optix::make_float3(res4.x, res4.y, res4.z);
}

RT_FUNCTION MediumSample call_medium_sample_distance(const SurfaceInteraction &si,
                                                    const optix::Ray &ray,
                                                    const optix::float2 &sample) {

    MediumSample mi;
    float unused_f;
    void *unused_ptr;

    detail::medium_sample_distance(
        // ----- Inputs
        si.t,
        // si.time
        0.f,
        // si.wavelengths
        0.f, 0.f, 0.f, 0.f,
        si.p.x, si.p.y, si.p.z,
        // si.shape
        nullptr,
        si.uv.x, si.uv.y,
        si.normal.x, si.normal.y, si.normal.z,
        si.sh_frame.s.x, si.sh_frame.s.y, si.sh_frame.s.z,
        si.sh_frame.t.x, si.sh_frame.t.y, si.sh_frame.t.z,
        si.sh_frame.n.x, si.sh_frame.n.y, si.sh_frame.n.z,
        // si.dp_du
        0.f, 0.f, 0.f,
        // si.dp_dv
        0.f, 0.f, 0.f,
        // si.duv_dx
        0.f, 0.f,
        // si.duv_dy
        0.f, 0.f,
        si.wi.x, si.wi.y, si.wi.z,
        // si.prim_index
        0,
        // si.instance
        nullptr,
        // ray origin and direction
        ray.origin.x, ray.origin.y, ray.origin.z,
        ray.direction.x, ray.direction.y, ray.direction.z,
        // ray.d_rcp
        1.f / ray.direction.x, 1.f / ray.direction.y, 1.f / ray.direction.z,
        ray.tmin, ray.tmax,
        // ray.time
        0.f,
        // ray.wavelengths
        0.f, 0.f, 0.f, 0.f,
        // sample
        sample.x, sample.y,
        // index (assumed unused!)
        0,
        // ----- Outputs
        &mi.t,
        &unused_f,  // mi.time
        &unused_f, &unused_f, &unused_f, &unused_f, // mi.wavelengths
        &mi.p.x, &mi.p.y, &mi.p.z,
        &unused_ptr, // mi.medium
        &mi.sh_frame.s.x, &mi.sh_frame.s.y, &mi.sh_frame.s.z,
        &mi.sh_frame.t.x, &mi.sh_frame.t.y, &mi.sh_frame.t.z,
        &mi.sh_frame.n.x, &mi.sh_frame.n.y, &mi.sh_frame.n.z,
        &mi.wi.x, &mi.wi.y, &mi.wi.z,

        &mi.throughput.x, &mi.throughput.y, &mi.throughput.z, &unused_f
    );
    return mi;
}

RT_FUNCTION float call_medium_pdf_distance(const optix::Ray &ray, float t) {
    float res;
    detail::medium_pdf_distance(
        ray.origin.x, ray.origin.y, ray.origin.z,
        ray.direction.x, ray.direction.y, ray.direction.z,
        // ray.d_rcp
        1.f / ray.direction.x, 1.f / ray.direction.y, 1.f / ray.direction.z,
        ray.tmin, ray.tmax,
        // ray.time
        0.f,
        // ray.wavelengths
        0.f, 0.f, 0.f, 0.f,
        // t
        t,
        // Output
        &res
    );
    return res;
}

RT_FUNCTION optix::float3 call_medium_eval_scattering(const MediumSample &mi) {
    optix::float4 res4;
    detail::medium_eval_scattering(
        mi.t,
        0.f,  // mi.time,
        0.f, 0.f, 0.f, 0.f,  // mi.wavelengths
        mi.p.x, mi.p.y, mi.p.z,
        nullptr,  // mi.medium
        mi.sh_frame.s.x, mi.sh_frame.s.y, mi.sh_frame.s.z,
        mi.sh_frame.t.x, mi.sh_frame.t.y, mi.sh_frame.t.z,
        mi.sh_frame.n.x, mi.sh_frame.n.y, mi.sh_frame.n.z,
        mi.wi.x, mi.wi.y, mi.wi.z,
        // Output
        &res4.x, &res4.y, &res4.z, &res4.w
    );
    return make_float3(res4.x, res4.y, res4.z);
}

RT_FUNCTION optix::float3 call_medium_eval_albedo(const MediumSample &mi) {
    optix::float4 res4;
    detail::medium_eval_albedo(
        mi.t,
        0.f,  // mi.time,
        0.f, 0.f, 0.f, 0.f,  // mi.wavelengths
        mi.p.x, mi.p.y, mi.p.z,
        // Output
        &res4.x, &res4.y, &res4.z, &res4.w
    );
    return make_float3(res4.x, res4.y, res4.z);
}

RT_FUNCTION float call_medium_density(float t, const optix::float3 &p) {
    float res;
    detail::medium_density(
        t,
        0.f,  // it.time,
        0.f, 0.f, 0.f, 0.f,  // it.wavelengths
        p.x, p.y, p.z,
        // Output
        &res
    );
    return res;
}
RT_FUNCTION float call_medium_density(const MediumSample &mi) {
    return call_medium_density(mi.t, mi.p);
}


RT_FUNCTION bool call_math_solve_quadratic(float a, float b, float c,
                                           float *solution1, float *solution2) {
    bool has_solution;
    detail::math_solve_quadratic(a, b, c, &has_solution, solution1, solution2);
    return has_solution;
}


/**
 * Returns a newly sampled direction in local coordinates.
 */
RT_FUNCTION optix::float3 sample_isotropic_phase(const optix::float2 &sample) {
    float z = 1.f - 2.f * sample.y;
    float r = std::sqrt(optix::fmaxf(1.f - z * z, 0.f));
    float theta = 2.f * math::Pi * sample.x;
    return optix::make_float3(r * std::cos(theta), r * std::sin(theta), z);
}

RT_FUNCTION float pdf_isotropic_phase() {
    return 0.25f * math::InvPi;
}

RT_FUNCTION float eval_isotropic_phase() {
    return pdf_isotropic_phase();
}

