#pragma once

#include <optix_world.h>
#include "common.cuh"
#include "media.cuh"

// Exposed by HeterogeneousMedium
rtDeclareVariable(float, medium_step_size, , );
rtDeclareVariable(optix::float3, medium_rgb_sigma_t, , );


RT_FUNCTION float medium_integrate_density(const optix::Ray &ray,
                                           float *out_last_interpolant = nullptr) {
    float mean_sigma_t = (medium_rgb_sigma_t.x + medium_rgb_sigma_t.y
                          + medium_rgb_sigma_t.z) / 3.f;

    float maxt = ray.tmax;
    float mint = ray.tmin;

    optix::float3 lookup_p;

    // TODO: consider adding jittering here
    float step_size = medium_step_size;
    int step_count  = 1 + (int) ceilf((maxt - mint) / step_size);

    float t_a  = mint;
    lookup_p   = ray.origin + t_a * ray.direction;
    float f_a  = call_medium_density(0.f, lookup_p) * mean_sigma_t;

    float integrated_density = 0.f;
    float last_interpolant = f_a;
    for (int i = 1; i < step_count; ++i) {
        float t_b  = fminf(fmaf(i, step_size, mint), maxt);
        lookup_p   = ray.origin + t_b * ray.direction;
        float f_b  = call_medium_density(0.f, lookup_p) * mean_sigma_t;

        last_interpolant = 0.5f * (f_a + f_b);
        integrated_density += last_interpolant * (t_b - t_a);
        t_a = t_b;
        f_a = f_b;
    }
    if (out_last_interpolant)
        *out_last_interpolant = last_interpolant;

    return integrated_density;
}
