#include <optix_world.h>
#include "common.cuh"
#include "media.cuh"
#include "media_d.cuh"
#include "medium_heterogeneous.cuh"

// Note: already declared in the header
// rtDeclareVariable(float, medium_step_size, , );
// rtDeclareVariable(optix::float3, medium_rgb_sigma_t, , );


namespace detail {

RT_FUNCTION void custom_eval_transmittance_d(const optix::Ray &ray,
                                             const optix::float3 &grad_of_result_) {
    // Density is actually scalar
    float grad_of_result = grad_of_result_.x + grad_of_result_.y
                           + grad_of_result_.z;

    float maxt      = ray.tmax;
    float mint      = ray.tmin;
    float step_size = medium_step_size;
    int step_count  = 1 + (int) ceilf((maxt - mint) / step_size);

    float mean_sigma_t = (medium_rgb_sigma_t.x + medium_rgb_sigma_t.y
                          + medium_rgb_sigma_t.z) / 3.f;
    float integrated_density = medium_integrate_density(ray);
    // TODO: could we do something to improve stability here?
    float weight = -expf(-integrated_density) * mean_sigma_t * grad_of_result;

    float t_a         = mint;
    optix::float3 p_a = ray.origin + t_a * ray.direction;

    for (int i = 1; i < step_count; ++i) {
        float t_b         = fminf(fmaf(i, step_size, mint), maxt);
        optix::float3 p_b = ray.origin + t_b * ray.direction;

        // TODO: combine these
        float this_weight = 0.5f * (t_b - t_a) * weight;
        // Contribution of f_a in this iteration
        call_medium_density_d(0.f, p_a, this_weight);
        // Contribution of f_b in this iteration
        call_medium_density_d(0.f, p_b, this_weight);
        t_a = t_b;
        p_a = p_b;
    }

}


}  // end namespace detail

extern "C" {

__device__ void medium_eval_transmittance_d(
    float ray_o_0, float ray_o_1, float ray_o_2,
    float ray_d_0, float ray_d_1, float ray_d_2,
    float /*ray_d_rcp_0*/, float /*ray_d_rcp_1*/, float /*ray_d_rcp_2*/,
    float ray_mint, float ray_maxt,
    float /*ray_time*/,
    float /*ray_wavelengths_0*/, float /*ray_wavelengths_1*/,
    float /*ray_wavelengths_2*/, float /*ray_wavelengths_3*/,
    uint32_t /* index */,
    float grad_of_result_0, float grad_of_result_1,
    float grad_of_result_2, float /*grad_of_result_3*/) {

    auto ray = optix::make_Ray(
        optix::make_float3(ray_o_0, ray_o_1, ray_o_2),
        optix::make_float3(ray_d_0, ray_d_1, ray_d_2),
        1,
        ray_mint, ray_maxt
    );
    detail::custom_eval_transmittance_d(
        ray,
        optix::make_float3(grad_of_result_0, grad_of_result_1, grad_of_result_2)
    );
}

}  // end extern "C"
