#pragma once

#include <optix_world.h>
#include <math_constants.h>
#include "types.cuh"
#include "common.cuh"
#include "envmap.cuh"

rtDeclareVariable(rtObject, top_object, , );
rtDeclareVariable(int, kHideEmitters, , "Whether to hide emitters from the final image (e.g. envmap)");

RT_FUNCTION RayPayload eval(size_t seed, optix::Ray ray, const float3 &delta,
                            uint max_depth, bool start_in_medium = false,
                            bool did_emitter_sampling = false) {
    RayPayload prd(seed, max_depth);
    prd.bsdf_origin = ray.origin;
    prd.bsdf_dir = ray.direction;
    prd.delta = delta;
    prd.in_medium = start_in_medium;
    prd.did_emitter_sampling = did_emitter_sampling;

    if (prd.max_depth <= 0)
        return prd;

    // Tracing inside of volume is done in hit kernel, so there's no looping here
    prd.escaped = true;
    rtTrace(top_object, ray, prd);

    if (!kHideEmitters && prd.escaped && !prd.did_emitter_sampling) {
        // Envmap is visible now
        prd.result += make_float4(prd.throughput, 0.f)
                      * make_float4(call_envmap_eval(prd.bsdf_dir), 0.f);
    }

    return prd;
}
