#include <mitsuba/optix/common.h>
#include <mitsuba/optix/renderer.h>
#include <mitsuba/python/python.h>
#include <mitsuba/render/bsdf.h>

#include "../ptx/types.cuh"

using namespace mitsuba;

MTS_PY_EXPORT(renderer) {
    py::class_<MyOptixState>(m, "MyOptixState");

    auto c =
        py::class_<OptiXRenderer, Object, ref<OptiXRenderer>>(m, "OptiXRenderer")
            .def(py::init<Scene *, OptiXRenderer::EIntegratorType, DifferentiableParameters *>(),
                 "scene"_a, "integrator"_a, "params"_a = nullptr)
            .def("init", &OptiXRenderer::init)
            .def("create_bsdfs_and_emitters_props", &OptiXRenderer::create_bsdfs_and_emitters_props)
            .def("extract_programs", &OptiXRenderer::extract_programs)
            .def("extract_bsdf_programs", &OptiXRenderer::extract_bsdf_programs)
            .def("extract_common_programs", &OptiXRenderer::extract_common_programs)
            .def("upload_bsdf_globals", &OptiXRenderer::upload_bsdf_globals)
            .def("create_emitter_props", &OptiXRenderer::create_emitter_props)
            .def("render", &OptiXRenderer::render, "seed"_a = 1, "spp"_a = (size_t) -1,
                 "pass_count"_a = 1)
            .def("delta_render",
                 py::overload_cast<size_t, const FloatC &, size_t, size_t, bool>(
                     &OptiXRenderer::delta_render),
                 "seed"_a, "image_derivatives"_a, "spp"_a = (size_t) -1, "pass_count"_a = 1,
                 "return_image"_a = false)
            .def("delta_render",
                 py::overload_cast<size_t, const Bitmap *, size_t, size_t, bool>(
                     &OptiXRenderer::delta_render),
                 "seed"_a, "image_derivatives"_a, "spp"_a = (size_t) -1, "pass_count"_a = 1,
                 "return_image"_a = false)
            .def("trace", &OptiXRenderer::trace)
            .def("unpack_gradients", &OptiXRenderer::unpack_gradients)
            // Optimization flags
            .def("use_approx_li_1", &OptiXRenderer::use_approx_li_1)
            .def("set_use_approx_li_1", &OptiXRenderer::set_use_approx_li_1);

    py::enum_<OptiXRenderer::EIntegratorType>(c, "EIntegratorType")
        .value("EDepth", OptiXRenderer::EDepth)
        .value("EDirect", OptiXRenderer::EDirect)
        .value("EPath", OptiXRenderer::EPath)
        .value("EVolPath", OptiXRenderer::EVolPath)
        .value("ERadiativeBackprop", OptiXRenderer::ERadiativeBackprop)
        .value("ERadiativeBackpropVol", OptiXRenderer::ERadiativeBackpropVol)
        .value("ERadiativeBackpropVolSimple", OptiXRenderer::ERadiativeBackpropVolSimple)
        .export_values();
}
