import os
from os.path import join, dirname, realpath

RES_SCENE_DIR = realpath(join(dirname(__file__), '../../../resources/data/scenes'))
DIFF_SCENE_DIR = realpath(join(dirname(__file__), '../../../../scenes'))
OUTPUT_DIR = realpath(join(DIFF_SCENE_DIR, '..', 'outputs'))
CBOX_PARAM_NAMES = [
    '/Scene/Mesh[id="_unnamed_10"]/BSDF[id="white_back"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_11"]/BSDF[id="green"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_12"]/BSDF[id="red"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_13"]/BSDF[id="white_small_box"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_14"]/BSDF[id="white_large_box"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_6"]/BSDF[id="white_luminaire"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_8"]/BSDF[id="white_floor"]/ContinuousSpectrum/rgb',
    '/Scene/Mesh[id="_unnamed_9"]/BSDF[id="white_ceiling"]/ContinuousSpectrum/rgb',
]
TEXTURED_PARAM_NAMES = [
    '/Scene/Mesh[id="_unnamed_13"]/BSDF[id="white_small_box"]/ContinuousSpectrum[id="_unnamed_14"]/bitmap',
    '/Scene/Mesh[id="_unnamed_15"]/BSDF[id="white_large_box"]/ContinuousSpectrum[id="_unnamed_16"]/bitmap',
]
HOMOGENEOUS_DENSITY_PARAM_NAMES = [
    '/Scene/Medium[id="medium1"]/density',
]
HETEROGENEOUS_DENSITY_PARAM_NAMES = [
    '/Scene/Medium[id="medium1"]/Texture3D[id="medium-density"]/data',
    # '/Scene/Medium[id="medium1"]/Texture3D[id="medium-density"]/ContinuousSpectrum/value',
]
ALBEDO_PARAM_NAMES = [
    '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_r',
    '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_g',
    '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_b',
]


CONFIGS = {
    'cbox-rgb': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb_initial.xml'),
        'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_rgb_ref.exr'),
        'params': CBOX_PARAM_NAMES,
        'scene_params': [('res', str(256)), ('max_depth', str(8))],
        'ref_scene_params': [('res', str(256)), ('max_depth', str(8))],
        'max_iterations': 100,
        'forward_spp': 32,
        'backward_spp': 32,
    },
    'cbox-rgb-approx': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb_initial.xml'),
        'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_rgb_ref.exr'),
        'params': CBOX_PARAM_NAMES,
        'scene_params': [('res', str(256)), ('max_depth', str(8))],
        'ref_scene_params': [('res', str(256)), ('max_depth', str(8))],
        'use_approx_li_1': True,
        'max_iterations': 100,
        'forward_spp': 32,
        'backward_spp': 32,
    },
    'cbox-rgb-md-2': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb_initial.xml'),
        'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_rgb.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_rgb_ref.exr'),
        'params': CBOX_PARAM_NAMES,
        'scene_params': [('res', str(256)), ('max_depth', str(2))],
        'ref_scene_params': [('res', str(256)), ('max_depth', str(2))],
        'forward_spp': 32,
        'backward_spp': 32,
    },

    'cbox-textured': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_textured.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_textured_ref.exr'),
        'params': TEXTURED_PARAM_NAMES,
        'scene_params': [('state', 'initial')],
        'ref_scene_params': [('state', 'target')],
        'learning_rate': 1e0,
    },
    'cbox-envmap': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_envmap.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_envmap_ref.exr'),
        'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_envmap_ref.xml'),
        'params': CBOX_PARAM_NAMES,
        'backward_spp': 32,
        'max_iterations': 100,
        'learning_rate': 1e-1,
        'ref_scene_params': [],
    },
    'cbox-envmap-only': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_envmap_only.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_envmap_only_ref.exr'),
        'ref_scene': join(RES_SCENE_DIR, 'cbox/cbox_envmap_only_ref.xml'),
        'params': CBOX_PARAM_NAMES,
        'backward_spp': 32,
        'ref_scene_params': [],
    },
    'cbox-microfacet': {
        'scene': join(RES_SCENE_DIR, 'cbox/cbox_microfacet.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/cbox_microfacet_ref.exr'),
        'params': [
            '/Scene/Mesh[id="_unnamed_8"]/BSDF[id="rough_floor"]/ContinuousSpectrum[id="rough_floor_alpha"]/value',
        ],
        'scene_params': [('roughness', '0.20')],
        'ref_scene_params': [('roughness', '0.05')],
    },

    'cbox-table': {
        'scene': join(DIFF_SCENE_DIR, 'cbox-table/cbox-table.xml'),
        'ref': join(DIFF_SCENE_DIR, 'cbox-table/cbox-table_ref.exr'),
        'params': ['/Scene/BSDF[id="TableWood"]/ContinuousSpectrum[id="diffuse_texture"]/bitmap'],
        'scene_params': [('state', 'initial')],
        'ref_scene_params': [('state', 'target')],
        'learning_rate': 1e-1,
        'max_iterations': 200,
        'backward_spp': 256,
    },


    'rb-debug-single': {
        'scene': join(DIFF_SCENE_DIR, '/rb-debug/debug_single.xml'),
        'ref': join(DIFF_SCENE_DIR, '/rb-debug/debug_single_small_ref.exr'),
        'params': [
            '/Scene/BSDF[id="red"]/ContinuousSpectrum/rgb',
        ],
    },


    'simpler': {
        'scene': join(RES_SCENE_DIR, 'cbox/simpler.xml'),
        'ref': join(RES_SCENE_DIR, 'cbox/simpler-ref.exr'),
        'params': [
            # '/Scene/Mesh[id="_unnamed_8"]/BSDF[id="white_floor"]/ContinuousSpectrum/rgb',
            # '/Scene/Mesh[id="_unnamed_8"]/BSDF[id="white_back"]/ContinuousSpectrum/rgb',
            '/Scene/Mesh[id="_unnamed_8"]/BSDF[id="red"]/ContinuousSpectrum/rgb',
        ],
    },
    'object-envmap-diffuse': {
        'scene': join(DIFF_SCENE_DIR, 'object-envmap/object-envmap.xml'),
        'ref': join(DIFF_SCENE_DIR, 'object-envmap/object-envmap-ref.exr'),
        'params': [
            '/Scene/Mesh[id="plane"]/BSDF[id="colored"]/ContinuousSpectrum/rgb',
        ],
        'forward_spp': 32,
        'backward_spp': 64,
        'scene_params': [
            ('max_depth', '2'),  # TODO
            ('color', '0.2, 0.4, 0.6'),
        ],
        'ref_scene_params': [
            ('max_depth', '2'),  # TODO
            ('color', '0.6, 0.2, 0.4'),
        ],
    },
    'object-envmap-rough': {
        'scene': join(DIFF_SCENE_DIR, 'object-envmap/object-envmap-rough.xml'),
        'ref': join(DIFF_SCENE_DIR, 'object-envmap/object-envmap-rough-ref.exr'),
        'params': [
            '/Scene/Mesh[id="plane"]/BSDF[id="microfacet"]/ContinuousSpectrum[id="microfacet_alpha"]/value',
        ],
        'fd_epsilon': 1e-3,
        'forward_spp': 32,
        'backward_spp': 64,  # 264,
        'scene_params': [
            ('max_depth', '2'),  # TODO
            ('roughness', '0.15'),
        ],
        'ref_scene_params': [
            ('max_depth', '2'),  # TODO
            ('roughness', '0.05'),
        ],
    },



    'rb-unit-absorptive': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-absorptive-homogeneous.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-absorptive_ref.exr'),
        'params': [
            # HOMOGENEOUS_DENSITY_PARAM_NAMES[0],
            '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/ContinuousSpectrum/value',
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        'scene_params': [('density', str(0.5))],
        # 'scene_params': [('density', str(1.5))],
        # 'scene_params': [('density', str(5))],
        # 'scene_params': [('density', str(10))],
        'ref_scene_params': [('density', str(2))],
        'max_iterations': 2,
        'forward_spp': 1024,
        # 'integrator': 'ERadiativeBackpropVolSimple',
    },

    'rb-unit-scattering-homogeneous': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-homogeneous.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-homogeneous_ref.exr'),
        'params': [
            '/Scene/Mesh[id="volume_bbox"]/Medium[id="medium1"]/density',
            # HOMOGENEOUS_DENSITY_PARAM_NAMES[0],
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/ContinuousSpectrum/value',
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        # 'scene_params': [('density', str(0.5))],
        # 'scene_params': [('density', str(1.5))],
        # 'scene_params': [('density', str(3))],
        'scene_params': [
            ('emitter_type', 'constant'),
            ('max_depth', str(2)),
            ('density', str(5))
        ],
        # 'scene_params': [('density', str(10))],
        'ref_scene_params': [('density', str(2))],
        'max_iterations': 3,
        # 'forward_spp': 32,
        # 'backward_spp': 32,
        'forward_spp': 1024,
        'backward_spp': 1024,
    },

    'rb-unit-scattering-heterogeneous': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-heterogeneous.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-heterogeneous_ref.exr'),
        'params': [
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/data',
            '/Scene/Mesh[id="volume_bbox"]/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        'is_heterogeneous': True,
        'scene_params': [
            ('emitter_type', 'envmap'),
            ('max_depth', str(8)),
            ('volume_filename', 'textures/constant-side-3-value-3.vol'),
        ],
        'ref_scene_params': [
            ('emitter_type', 'envmap'),
            ('max_depth', str(8)),
            ('volume_filename', 'textures/noise-side-3.vol'),
        ],
        'max_iterations': 10,
        'forward_spp': 128,
        'backward_spp': 128,
    },

    'rb-unit-full': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-full.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-full_ref.exr'),
        'params': [
            # HOMOGENEOUS_DENSITY_PARAM_NAMES[0],
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/ContinuousSpectrum/value',
            '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        'scene_params': [('density', str(0.5))],
        # 'scene_params': [('density', str(1.5))],
        # 'scene_params': [('density', str(3))],
        # 'scene_params': [('density', str(5))],
        # 'scene_params': [('density', str(10))],
        'ref_scene_params': [('density', str(2))],
        'max_iterations': 3,
        # 'forward_spp': 32,
        # 'backward_spp': 32,
        # 'forward_spp': 64,
        # 'backward_spp': 64,
        'forward_spp': 1024,
        'backward_spp': 1024,
    },

    'rb-unit-scattering-hide-emitters': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-unit-scattering-hide-emitters_ref.exr'),
        'params': [
            HOMOGENEOUS_DENSITY_PARAM_NAMES[0],
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/ContinuousSpectrum/value',
            # '/Scene/Medium[id="medium1"]/Texture3D[id="density"]/data',
        ],
        'has_volumes': True,
        # 'scene_params': [('density', str(3)), ('hide_emitters', 'false')],
        'scene_params': [('density', str(10)), ('hide_emitters', 'true')],
        'ref_scene_params': [('density', str(2)), ('hide_emitters', 'true')],
        'max_iterations': 3,
        # 'forward_spp': 1,
        # 'backward_spp': 1,
        # 'forward_spp': 32,
        # 'backward_spp': 32,
        'forward_spp': 1024,
        'backward_spp': 1024,
    },

    'rb-unit-albedo': {
        'scene': join(DIFF_SCENE_DIR, 'albedo-opt/rb-unit-albedo.xml'),
        'ref': join(DIFF_SCENE_DIR, 'albedo-opt/rb-unit-albedo_ref.exr'),
        'params': [
            '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_r',
            '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_g',
            '/Scene/Medium[id="medium1"]/Texture3D[id="albedo"]/data_b',
        ],
        'has_volumes': True,
        'scene_params': [('albedo_filename', 'textures/albedo-grid-side-8-constant-0.75.vol')],
        # 'scene_params': [('albedo_filename', 'textures/albedo-grid-side-8-constant-red.vol')],
        'ref_scene_params': [('albedo_filename', 'textures/albedo-grid-side-8-gradient.vol')],
        'max_iterations': 100,
        'forward_spp': 64,
        'backward_spp': 64,
    },


    'rb-homogeneous-noise-opt': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-homogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-homogeneous-opt_ref.exr'),
        'params': HOMOGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'scene_params': [('density', str(0.5))],
        'ref_scene_params': [('density', str(2))],
    },
    'rb-homogeneous-noise-opt-hidden-emitters': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-homogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-homogeneous-opt-hidden-emitters_ref.exr'),
        'params': HOMOGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'scene_params': [('density', str(0.5)), ('hide_emitters', 'true')],
        'ref_scene_params': [('density', str(2)), ('hide_emitters', 'true')],
    },
    'rb-heterogeneous-noise-opt': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt_ref.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': True,
        'scene_params': [('volume_filename', 'textures/constant-side-3-value-0.1.vol')],
        'ref_scene_params': [('volume_filename', 'textures/noise.vol')],
        'forward_spp': 256,
        'backward_spp': 256,
    },

    # 'smoke-plume': {
    #     'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
    #     'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt_ref.exr'),
    #     'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
    #     'has_volumes': True,
    #     'is_heterogeneous': True,
    #     'scene_params': [('volume_filename', 'textures/smoke.vol')]
    # },

    'smoke-plume': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt_ref.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': True,
        'forward_spp': 128,
        'backward_spp': 128,
        'scene_params': [('volume_filename', 'textures/density-grid-res-111-222-112-value-0.018.vol')],
        'ref_scene_params': [('volume_filename', 'textures/smoke.vol')],
    },

    'smoke-plume-approx': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt_ref.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'use_approx_li_1': True,  # Approximate gradients
        'has_volumes': True,
        'is_heterogeneous': True,
        'forward_spp': 128,
        'backward_spp': 128,
        'scene_params': [('volume_filename', 'textures/density-grid-res-111-222-112-value-0.018.vol')],
        'ref_scene_params': [('volume_filename', 'textures/smoke.vol')],
    },

    'smoke-plume-res-111-white': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/smoke-plume_ref-white.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': True,
        # 'forward_spp': 64,
        # 'backward_spp': 64,
        'forward_spp': 256,
        'backward_spp': 256,
        'learning_rate': 2e-2, #2e0,  #1e0,
        'max_iterations': 300,
        'scene_params': [
            ('res', str(256)),
            ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.1.vol'),
            # ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.018.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
        ],
        'ref_scene_params': [
            ('res', str(256)),
            ('volume_filename', 'textures/smoke.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
        ],
    },

    'smoke-plume-res-111-white-approx': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/smoke-plume_ref-white.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': True,
        'use_approx_li_1': True,  # Approximate gradients
        'forward_spp': 256,
        'backward_spp': 256,
        'learning_rate': 2e-2,
        'max_iterations': 300,
        'scene_params': [
            ('res', str(256)),
            ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.1.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
        ],
        'ref_scene_params': [
            ('res', str(256)),
            ('volume_filename', 'textures/smoke.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
        ],
    },

    'smoke-plume-res-111-white-hide': {
        'scene': join(DIFF_SCENE_DIR, 'medium-opt/rb-heterogeneous-opt.xml'),
        'ref': join(DIFF_SCENE_DIR, 'medium-opt/smoke-plume_ref-white-hide.exr'),
        'params': HETEROGENEOUS_DENSITY_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': True,
        'forward_spp': 512,
        'backward_spp': 512,
        'learning_rate': 2e-2, #2e0,  #1e0,
        'max_iterations': 300,
        'scene_params': [
            ('volume_filename', 'textures/density-grid-res-111-222-112-value-0.1.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
            ('hide_emitters', 'true'),
        ],
        'ref_scene_params': [
            ('volume_filename', 'textures/smoke.vol'),
            ('smoke_color', '1.0, 1.0, 1.0'),
            ('hide_emitters', 'true'),
        ],
    },

    'albedo-opt-null': {
        'scene': join(DIFF_SCENE_DIR, 'albedo-opt/rb-albedo-opt.xml'),
        # 'ref': join(DIFF_SCENE_DIR, 'albedo-opt/rb-albedo-opt_ref.exr'),
        'ref': join(DIFF_SCENE_DIR, 'albedo-opt/colortargets/van-gogh-256.exr'),
        'params': ALBEDO_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': False,
        # 'max_iterations': 100,
        'forward_spp': 64,
        'backward_spp': 64,
        # 'scene_params': [('initial_value', 'textures/constant-side-2-value-0.018.vol')],
        # 'scene_params': [('initial_value', 'textures/albedo-grid-side-8-constant-0.75.vol')],
        'scene_params': [
            ('res', '256'),
            ('density_scale', '15'),
            ('initial_value', 'textures/albedo-256-64-256-value-0.75.vol')
        ],
        'ref_scene_params': [
            ('res', '256'),
            ('density_scale', '15'),
            ('initial_value', 'colortargets/van-gogh-256.vol')
        ],
    },
    'albedo-opt-null-approx': {
        'scene': join(DIFF_SCENE_DIR, 'albedo-opt/rb-albedo-opt.xml'),
        # 'ref': join(DIFF_SCENE_DIR, 'albedo-opt/rb-albedo-opt_ref.exr'),
        'ref': join(DIFF_SCENE_DIR, 'albedo-opt/colortargets/van-gogh-256.exr'),
        'params': ALBEDO_PARAM_NAMES,
        'has_volumes': True,
        'is_heterogeneous': False,
        'use_approx_li_1': True,
        # 'max_iterations': 100,
        'forward_spp': 64,
        'backward_spp': 64,
        # 'scene_params': [('initial_value', 'textures/constant-side-2-value-0.018.vol')],
        # 'scene_params': [('initial_value', 'textures/albedo-grid-side-8-constant-0.75.vol')],
        'scene_params': [
            ('res', '256'),
            ('density_scale', '15'),
            ('initial_value', 'textures/albedo-256-64-256-value-0.75.vol')
        ],
        'ref_scene_params': [
            ('res', '256'),
            ('density_scale', '15'),
            ('initial_value', 'colortargets/van-gogh-256.vol')
        ],
    },


    'bell': {
        'scene': join(DIFF_SCENE_DIR, 'bell/bell.xml'),
        'scene_lowres': join(DIFF_SCENE_DIR, 'bell/bell_lowres.xml'),
        'ref': join(DIFF_SCENE_DIR, 'bell/bell_ref.exr'),
        'params': [
            '/Scene/BSDF[id="BC_sphere"]/ContinuousSpectrum[id="texture_earth"]/bitmap',
        ],
        'scene_params': [('state', 'initial')],
        'ref_scene_params': [('state', 'target')],
        'max_iterations': 160,
        # 'learning_rate': 8.0e7,# SGD:
        # 'learning_rate': 2.4e1,# ADAM
        'learning_rate': 0.5e1,# ADAM
        'block_size': 256,
        # 'use_approx_li_1': True,
        'forward_spp': 32,
        'backward_spp': 32,
        'ref_spp': 1024,
        'save_intermediate_results': True,
    },

    'no_bell': {
        'scene': join(DIFF_SCENE_DIR, 'bell/no_bell.xml'),
        'ref': join(DIFF_SCENE_DIR, 'bell/no_bell_scene_ref.exr'),
        'params': [
            '/Scene/BSDF[id="BC_sphere"]/ContinuousSpectrum[id="texture_earth"]/bitmap',
        ],
        'scene_params': [('state', 'initial')],
        'ref_scene_params': [('state', 'target')],
        'max_iterations': 100,
        'learning_rate': 3e-2,
        'block_size': 256,
        'use_approx_li_1': True,
        'forward_spp': 32,
        'backward_spp': 32,
        'ref_spp': 1024,
        'save_intermediate_results': True,
    },
}
