import os
from os.path import realpath, join

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from config import OUTPUT_DIR


def plot_one(data, fig, ax, legend=None):
    ax[0].plot(data['iteration'], data['loss'], label=legend)
    ax[0].set_xlabel('Iteration')

    cumulated = [0] + np.cumsum(data['runtime'])
    ax[1].plot(cumulated, data['loss'])
    ax[1].set_xlabel('Time (s)')


def main():
    output_dir = join(OUTPUT_DIR, 'cbox-textured')

    files = [
        # join(output_dir, 'metadata-single.csv'),
        join(output_dir, 'metadata.csv'),
    ]
    legends = [
        "Off-by-one (single pass)",
        "Correct version (two passes)",
    ]

    fig, ax = plt.subplots(2, 1, figsize=(12, 2 * 8))
    for fname, legend in zip(files, legends):
        df = pd.read_csv(fname, sep=',', header=0, index_col=False,
                         squeeze=True, skipinitialspace=True)
        plot_one(df, fig, ax, legend=legend)

    fig.legend()
    plt.tight_layout()
    plt.show()
    # out = join(OUTPUT_DIR, filename)
    # plt.savefig(out, bbox_inches='tight', dpi=128)
    # plt.savefig(splitext(out)[0] + '.png', bbox_inches='tight', dpi=200)

if __name__ == '__main__':
    main()
