import numpy as np


def add_colorbar(fig, ax, img_widget, **kwargs):
    """Source: https://stackoverflow.com/a/39938019/3792942"""
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)

    return fig.colorbar(img_widget, cax=cax, orientation='vertical', **kwargs)


def add_outlined_text(ax, position, text, size=14):
    import matplotlib.patheffects as path_effects

    text = ax.text(position[0], position[1], text,
                   color='white', ha='left', va='center', size=size)
    text.set_path_effects([path_effects.Stroke(linewidth=3, foreground='black'),
                           path_effects.Normal()])
    return text


def add_bar_labels(ax, rectangles, values = None, ylims = None, label_format = '{:.3e}',
                   alternate_color=None, **kwargs):
    """Adds a vertical text label to each bar with its value."""
    heights = [rect.get_height() for rect in rectangles]
    max_height = ylims[-1] if ylims else np.max(heights)
    values = values or heights
    for i, average in enumerate(values):
        rect   = rectangles[i]
        y_text = heights[i] + 0.015 * max_height
        color = 'black'
        if y_text >= 0.80 * max_height:
            y_text = min(max_height, heights[i]) - 0.10 * max_height
            color = alternate_color or color

        # Bar in the same direction
        ax.text(rect.get_x() + 0.56 * rect.get_width(), y_text,
                label_format.format(average), ha='center', va='bottom', rotation=90,
                color=color, **kwargs)
