import os
from os.path import join, dirname, realpath
import platform

from config import CONFIGS, OUTPUT_DIR

from render_helpers import save_info_csv

def creation_date(path_to_file):
    """
    Try to get the date that a file was created, falling back to when it was
    last modified if that isn't possible.
    See http://stackoverflow.com/a/39501288/1709587 for explanation.
    """
    if platform.system() == 'Windows':
        return os.path.getctime(path_to_file)
    else:
        stat = os.stat(path_to_file)
        try:
            return stat.st_birthtime
        except AttributeError:
            # We're probably on Linux. No easy way to get creation dates here,
            # so we'll settle for when its content was last modified.
            return stat.st_mtime


if __name__ == '__main__':

    config_name = 'bell'

    config = CONFIGS[config_name]
    output_dir = join(OUTPUT_DIR, config_name)

    prefix = 'ad'

    max_iterations = config.get('max_iterations', 150)

    start_time = creation_date(join(output_dir, 'ref.exr'))

    runtime = []

    for it_i in range(max_iterations):
        texture_fname = join(output_dir, prefix + '_opt_texture_{:04d}.exr'.format(it_i))

        if not os.path.isfile(texture_fname):
            break

        end_time = creation_date(texture_fname)

        # print(end_time - start_time)
        runtime.append(end_time - start_time)

        start_time = end_time


    fname = join(output_dir, prefix + '_metadata_runtime_recreated.csv')
    save_info_csv(fname, iteration=list(range(len(runtime))), runtime=runtime)
    print('[+] Saved run metadata to: {}'.format(fname))

