import os
from os.path import join, dirname, realpath
import pytest
import time

import mitsuba
from mitsuba.core import Bitmap, Struct, Thread, EColorMode
from mitsuba.optix import OptiXRenderer
import numpy as np

from render_helpers import load_scene, integrator_type_for_scene
from config import OUTPUT_DIR, RES_SCENE_DIR


def test01_change_sensor():
    output_dir = join(OUTPUT_DIR, 'cbox-multiview')
    os.makedirs(output_dir, exist_ok=True)

    scene_fname = join(RES_SCENE_DIR, 'cbox', 'cbox_multiview.xml')
    scene = load_scene(scene_fname)
    assert len(scene.sensors()) == 2

    renderer = OptiXRenderer(scene, integrator_type_for_scene(scene, with_gradients=False))
    for i, _ in enumerate(scene.sensors()):
        print('[ ] Rendering from sensor {}'.format(i))
        scene.set_current_sensor(i)
        current_b = renderer.render(seed=1234, spp=64)

        fname = join(output_dir, 'sensor_{:02d}.exr'.format(i))
        current_b.write(fname)
        print('[ ] Saved sensor {} image to: {}'.format(i, fname))


if __name__ == '__main__':
    test01_change_sensor()
