import gc
import threading
import time

import enoki as ek
import numpy as np


def work():
    # 1. Buffer
    ek.cuda_malloc_trim()
    time.sleep(0.25)

    # 2. Alloc 1
    values = ek.FloatC.zero(int(1e8))
    ek.cuda_eval()
    time.sleep(0.5)

    # 3. Alloc 2
    values2 = ek.FloatC.zero(int(1e9))
    ek.cuda_eval()
    time.sleep(0.5)

    # 4. Cleanup
    del values, values2
    gc.collect()
    ek.cuda_malloc_trim()
    time.sleep(0.5)


def used_gpu_ram():
    free, total = ek.cuda_mem_get_info()
    return total - free


def plot_usage(times, used):
    import matplotlib.pyplot as plt
    _, ax = plt.subplots(1, 1)
    ax.plot(np.array(times) - times[0], used)
    # ax.set_ylim((0, ax.get_ylim()[1]))
    plt.show()


def test01_plot_memory_usage():
    th = threading.Thread(target=work)
    th.start()

    times = []
    used = []
    times.append(time.time())
    used.append(used_gpu_ram())

    while th.is_alive():
        times.append(time.time())
        used.append(used_gpu_ram())
        th.join(timeout=1e-2)
    th.join()

    time.sleep(1)
    times.append(time.time())
    used.append(used_gpu_ram())


    print('Complete, made {} RAM measurements'.format(len(used)))
    plot_usage(times, used)


def test02_sampler_thread():
    from .benchmark_utils import GPUUsageSampler
    sampler = GPUUsageSampler()

    sampler.start()
    work()
    times, used = sampler.done()


    # Deduct max usage of `work` from data
    print('\nMedian:      ', sampler.median_usage())
    print('Max:         ', sampler.max_usage())
    expected = int((1e8 + 1e9) * 4)
    print('Expected:    ', expected)
    print('Starting:    ', used[0])
    print('Middle:      ', used[len(used) // 2] - used[0])

    assert np.allclose(sampler.max_usage(), expected, rtol=0.1)
    plot_usage(times, used)


if __name__ == '__main__':
    test01_plot_memory_usage()
