from collections import namedtuple

KEY_FIELDS = ['scene_name', 'resolution_x', 'resolution_y', 'spp', 'max_depth']
RESULT_FIELDS = ['loss', 'time_per_it', 'total_time', 'iteration_count', 'memory_usage', 'gpu_usage']
FIELD_TYPES = {
    'scene_name': str,
    'resolution_x': int,
    'resolution_y': int,
    'spp': int,
    'max_depth': int,
    'loss': float,
    'time_per_it': float,   # In seconds
    'total_time': float,    # In seconds
    'iteration_count': int,
    'memory_usage': float,  # In bytes
    'gpu_usage': float,     # In percentage
}
assert set(FIELD_TYPES.keys()) == set(KEY_FIELDS + RESULT_FIELDS)

ResultKey = namedtuple('ResultKey', KEY_FIELDS)
Result = namedtuple('Result', ['key'] + RESULT_FIELDS)


def make_key(**kwargs):
    cast_args = { k: (FIELD_TYPES[k](v) if v is not None else None)
                  for k, v in kwargs.items() }
    return ResultKey(**cast_args)


def key_to_string(method_name, key, exclude_spp=False):
    extra = '' if exclude_spp else '_{spp}spp'.format(spp=key.spp)
    fmt = '{scene_name}_{method_name}_{resolution_x}x{resolution_y}{extra}_{max_depth}md'
    return fmt.format(method_name=method_name, extra=extra, **key._asdict())


def scene_params_for_key(key, config, is_reference):
    params = [('resx', str(key.resolution_x)), ('resy', str(key.resolution_y)),
              ('res', str(key.resolution_x)),  # In case it's assumed square
              ('spp', str(key.spp)),
              ('max_depth', str(key.max_depth))]
    extra_params_key = 'ref_scene_params' if is_reference else 'scene_params'
    params += config.get(extra_params_key, [])
    return params
