import numpy as np


def write_binary_grid3d(filename, values):
    with open(filename, 'wb') as f:
        f.write(b'V')
        f.write(b'O')
        f.write(b'L')
        f.write(np.uint8(3).tobytes())  # Version
        f.write(np.int32(1).tobytes())  # type
        f.write(np.int32(values.shape[0]).tobytes())  # size
        f.write(np.int32(values.shape[1]).tobytes())
        f.write(np.int32(values.shape[2]).tobytes())
        if values.ndim == 3:
            f.write(np.int32(1).tobytes())  # channels
        else:
            f.write(np.int32(values.shape[3]).tobytes())  # channels
        f.write(np.float32(0.0).tobytes())  # bbox
        f.write(np.float32(0.0).tobytes())
        f.write(np.float32(0.0).tobytes())
        f.write(np.float32(1.0).tobytes())
        f.write(np.float32(1.0).tobytes())
        f.write(np.float32(1.0).tobytes())
        f.write(values.ravel().astype(np.float32).tobytes())


def read_binary_grid3d(filename):
    end = 'little'
    dt = np.dtype(np.float32)
    dt = dt.newbyteorder('<')

    with open(filename, 'rb') as f:
        if f.read(3) != b'VOL':
            raise ValueError(f"File {filename} is not a Mitsuba volume file!")
        version = int.from_bytes(f.read(1), byteorder=end)
        file_type = int.from_bytes(f.read(4), byteorder=end)
        size0 = int.from_bytes(f.read(4), byteorder=end)
        size1 = int.from_bytes(f.read(4), byteorder=end)
        size2 = int.from_bytes(f.read(4), byteorder=end)
        channels = int.from_bytes(f.read(4), byteorder=end)

        bbox_x_min = np.frombuffer(f.read(4), dtype=dt)
        bbox_y_min = np.frombuffer(f.read(4), dtype=dt)
        bbox_z_min = np.frombuffer(f.read(4), dtype=dt)
        bbox_x_max = np.frombuffer(f.read(4), dtype=dt)
        bbox_y_max = np.frombuffer(f.read(4), dtype=dt)
        bbox_z_max = np.frombuffer(f.read(4), dtype=dt)
        values = np.frombuffer(f.read(4 * size0 * size1 * size2 * channels), dtype=dt)

        if channels == 1:
            return np.reshape(values, (size0, size1, size2))
        else:
            return np.reshape(values, (size0, size1, size2, channels))
