#include <mitsuba/core/properties.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/render/phase.h>
#include <mitsuba/render/spectrum.h>

NAMESPACE_BEGIN(mitsuba)

class IsotropicPhaseFunction final : public PhaseFunction {
public:
    IsotropicPhaseFunction(const Properties & /* props */) {}

    template <typename MediumInteraction, typename Point2,
              typename Value   = typename Point2::Value,
              typename Vector3 = typename MediumInteraction::Vector3>
    MTS_INLINE Vector3 sample_impl(const MediumInteraction & /*mi*/,
                                   const Point2 &sample1,
                                   mask_t<Value> /*active*/) const {
        return warp::square_to_uniform_sphere(sample1);
    }

    template <typename Vector3, typename Value = value_t<Vector3>>
    MTS_INLINE Value eval_impl(const Vector3 &wo,
                               mask_t<Value> /*active*/) const {
        return warp::square_to_uniform_sphere_pdf(wo);
    }

    std::string to_string() const override {
        return "IsotropicPhaseFunction[]";
    }

    MTS_IMPLEMENT_PHASE_FUNCTION_ALL()
    MTS_DECLARE_CLASS()
private:
};

MTS_IMPLEMENT_CLASS(IsotropicPhaseFunction, PhaseFunction)
MTS_EXPORT_PLUGIN(IsotropicPhaseFunction, "Isotropic phase function")

NAMESPACE_END(mitsuba)
