import enoki as ek
import numpy as np

from mitsuba.core import Bitmap, Struct
from mitsuba.render import ImageBlock


def image_to_float_d(image, pixel_format):
    W = None
    if isinstance(image, ImageBlock):
        X, Y, Z, A, W = image.bitmap_d()
    elif isinstance(image, Bitmap):

        channels = []
        arr = np.array(image, copy=False)
        for i in range(image.channel_count()):
            channels.append(ek.FloatD(arr[:, :, i].flatten()))
        if len(channels) == 1:
            assert pixel_format == Bitmap.EY
            assert image.pixel_format() == pixel_format
            return channels[0]
        elif len(channels) == 3:
            assert pixel_format in (Bitmap.ERGB, Bitmap.EXYZ)
            assert image.pixel_format() == pixel_format
            return ek.Vector3fD(channels[0], channels[1], channels[2])
        elif len(channels) == 4:
            assert pixel_format == Bitmap.ERGBA
            assert image.pixel_format() == pixel_format
            return ek.Vector4fD(channels[0], channels[1], channels[2], channels[3])
        else:
            assert len(channels) == 5
            X, Y, Z, A, W = channels
    else:
        raise RuntimeError('Cannot convert image of type: {}'.format(type(image)))

    if W is not None:
        W = ek.select(ek.eq(W, 0), 0, ek.rcp(W))
        X *= W
        Y *= W
        Z *= W
        A *= W

    if pixel_format == Bitmap.EY:
        return Y
    elif pixel_format == Bitmap.EXYZ:
        return ek.Vector3fD(X, Y, Z)
    elif pixel_format == Bitmap.ERGB:
        return ek.Vector3fD(
            3.240479 * X + -1.537150 * Y + -0.498535 * Z,
            -0.969256 * X + 1.875991 * Y + 0.041556 * Z,
            0.055648 * X + -0.204043 * Y + 1.057311 * Z
        )
    elif pixel_format == Bitmap.ERGBA:
        return ek.Vector4fD(
            3.240479 * X + -1.537150 * Y + -0.498535 * Z,
            -0.969256 * X + 1.875991 * Y + 0.041556 * Z,
            0.055648 * X + -0.204043 * Y + 1.057311 * Z,
            A
        )
    else:
        raise RuntimeError('Unsupported pixel format: {}'.format(pixel_format))


def float_d_to_bitmap(values, shape, pixel_format):
    return Bitmap(values.numpy().reshape(*shape), pixel_format=pixel_format)
