#include <mitsuba/render/spectrum.h>
#include <mitsuba/render/interaction.h>
#include <mitsuba/core/properties.h>

NAMESPACE_BEGIN(mitsuba)

class ProductSpectrum final : public ContinuousSpectrum {
public:
    ProductSpectrum(const Properties &props) : ContinuousSpectrum(props) {
        m_spectrum1 = props.spectrum("spectrum1");
        m_spectrum2 = props.spectrum("spectrum2");
    }

    template <typename Value, typename Mask = mask_t<Value>>
    MTS_INLINE Value eval_impl(Value wavelengths, Mask active) const {
        return m_spectrum1->eval(wavelengths, active) * m_spectrum2->eval(wavelengths, active);
    }

    std::vector<ref<Object>> children() override {
        if (m_spectrum1->id().empty())
            m_spectrum1->set_id(this->id() + "_spectrum1");
        if (m_spectrum2->id().empty())
            m_spectrum2->set_id(this->id() + "_spectrum2");
        return { m_spectrum1.get(), m_spectrum2.get() };
    }

    std::string to_string() const override {
        std::ostringstream oss;
        oss << "SumSpectrum[" << std::endl
            << "  spectrum1 = " << string::indent(m_spectrum1->to_string()) << "," << std::endl
            << "  spectrum2 = " << string::indent(m_spectrum2->to_string()) << std::endl
            << "]";
        return oss.str();
    }

#if defined(MTS_ENABLE_AUTODIFF)
    void set_labels(const std::string &prefix) const override {
        m_spectrum1->set_labels(prefix + ".m_spectrum1");
        m_spectrum2->set_labels(prefix + ".m_spectrum2");
    }

    void set_globals(const std::string &prefix, const RTcontext &context,
                     const RTmaterial &material) const override {
        m_spectrum1->set_globals(prefix + "_m_spectrum1", context, material);
        m_spectrum2->set_globals(prefix + "_m_spectrum2", context, material);
    }
#endif

    MTS_IMPLEMENT_SPECTRUM_EVAL_ALL()
    MTS_DECLARE_CLASS()

private:
    ref<ContinuousSpectrum> m_spectrum1, m_spectrum2;
};

MTS_IMPLEMENT_CLASS(ProductSpectrum, ContinuousSpectrum)
MTS_EXPORT_PLUGIN(ProductSpectrum, "Product spectrum")

NAMESPACE_END(mitsuba)
